
/**
  * Author: Jan Fuchs
  * $Date: 2009-04-02 14:10:26 +0200 (Thu, 02 Apr 2009) $
  * $Rev: 8 $
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <semaphore.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include <sys/sem.h>

#include "global.h"
#include "ipc.h"

PIX *p_pix_sm;
CCBUF *p_ccbuf_sm;

int pix_shmid;
int pix_semid;
int ccbuf_shmid;
int ccbuf_semid;

void lock(
    int semid,
    int num);
void unlock(
    int semid,
    int num);
void ipc_create(
    void);
void ipc_remove(
    void);

static void set_semvalue(
    int semid,
    int num);
static void del_semvalue(
    int semid,
    int num);

/**
  *
 */
static void
set_semvalue(
    int semid,
    int num)
{
    union semun sem_union;

    sem_union.val = 1;

    if (semctl(semid, num, SETVAL, sem_union) == -1) {
        syslog(LOG_ERR, "Failed to initialize semaphore %i\n", semid);
        exit(EXIT_FAILURE);
    }
}

/**
  *
 */
static void
del_semvalue(
    int semid,
    int num)
{
    union semun sem_union;

    if (semctl(semid, num, IPC_RMID, sem_union) == -1) {
        syslog(LOG_ERR, "Failed to delete semaphore\n");
    }
}

/**
  *
 */
void
lock(
    int semid,
    int num)
{
    struct sembuf sem_b;

    sem_b.sem_num = num;
    sem_b.sem_op = -1;
    sem_b.sem_flg = SEM_UNDO;

    if (semop(semid, &sem_b, 1) == -1) {
        syslog(LOG_ERR, "semaphore_p failed\n");
        exit(EXIT_FAILURE);
    }
}

/**
  *
 */
void
unlock(
    int semid,
    int num)
{
    struct sembuf sem_b;

    sem_b.sem_num = num;
    sem_b.sem_op = 1;
    sem_b.sem_flg = SEM_UNDO;

    if (semop(semid, &sem_b, 1) == -1) {
        syslog(LOG_ERR, "semaphore_v failed\n");
        exit(EXIT_FAILURE);
    }
}

/**
  *
 */
void
ipc_create(
    void)
{
    pix_shmid = shmget((key_t) 1234, sizeof(PIX), 0666 | IPC_CREAT);
    if (pix_shmid == -1) {
        syslog(LOG_ERR, "shmget failed\n");
        exit(EXIT_FAILURE);
    }

    p_pix_sm = (PIX *) shmat(pix_shmid, (void *) 0, 0);
    if ((void *) p_pix_sm == (void *) -1) {
        syslog(LOG_ERR, "shmat failed\n");
        exit(EXIT_FAILURE);
    }

    ccbuf_shmid = shmget((key_t) 4321, sizeof(CCBUF), 0666 | IPC_CREAT);
    if (ccbuf_shmid == -1) {
        syslog(LOG_ERR, "shmget failed ccbuf\n");
        exit(EXIT_FAILURE);
    }

    p_ccbuf_sm = (CCBUF *) shmat(ccbuf_shmid, (void *) 0, 0);
    if ((void *) p_ccbuf_sm == (void *) -1) {
        syslog(LOG_ERR, "shmat failed ccbuf\n");
        exit(EXIT_FAILURE);
    }

    pix_semid = semget((key_t) 1234, 1, 0666 | IPC_CREAT);
    set_semvalue(pix_semid, 0);

    ccbuf_semid = semget((key_t) 1235, 1, 0666 | IPC_CREAT);
    set_semvalue(ccbuf_semid, 0);
}

/**
  *
 */
void
ipc_remove(
    void)
{
    if (shmdt((void *) p_pix_sm) == -1) {
        syslog(LOG_ERR, "shmdt failed\n");
    }

    if (shmctl(pix_shmid, IPC_RMID, 0) == -1) {
        syslog(LOG_ERR, "shmctl(IPC_RMID) failed\n");
    }

    if (shmdt((void *) p_ccbuf_sm) == -1) {
        syslog(LOG_ERR, "shmdt failed\n");
    }

    if (shmctl(ccbuf_shmid, IPC_RMID, 0) == -1) {
        syslog(LOG_ERR, "shmctl(IPC_RMID) failed\n");
    }

    del_semvalue(pix_semid, 0);
    del_semvalue(ccbuf_semid, 0);
}
