
/**
  * Author: Jan Fuchs
  * $Date: 2009-04-02 14:10:26 +0200 (Thu, 02 Apr 2009) $
  * $Rev: 8 $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <syslog.h>
#include <getopt.h>
#include <time.h>
#include <malloc.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include <sys/sem.h>
#include <GL/freeglut.h>
//#include <linux/videodev.h>
#include <asm/types.h>
#include <linux/videodev2.h>

#include "global.h"
#include "ipc.h"
#include "napoveda.h"

#define CHANNEL_TUNER     0
#define CHANNEL_COMPOSITE 1
#define CHANNEL_SVIDEO    2

#define VDEPTH  3
#define V4L_DEVICE_MAX 127

#define CLEAR(x) memset (&(x), 0, sizeof (x))

struct buffer {
    void *start;
    size_t length;
};

typedef struct {
    int test_mode;
    int v4l_input;
    int v4l_standard;
    char v4l_device[V4L_DEVICE_MAX + 1];
    unsigned long usleep;
    IO_METHOD_T io_method;
} USER_SETTINGS_T;

static int fd = -1;
struct buffer *buffers = NULL;
static unsigned int n_buffers = 0;

/* Deprecate API V4L1 */
//struct vdev {
//  int video_dev;
//  int width;
//  int height;
//  int input;
//  int format;
//  struct video_channel video_chan;
//  struct video_tuner tuner;
//  struct video_mmap vid_mmap[2];
//  struct video_mbuf vid_mbuf;
//  char *p_video_map;
//};

struct {
    int run;
    long seek;
    FILE *p_fr;
} impo;

struct {
    char *p_soubor;
    FILE *p_fw;
} expo;

struct {
    int width;
    int height;
    int x;
    int y;
} window;

char dev_video[V4L_DEVICE_MAX + 1];

PIX pix;

/* Deprecate API V4L1 */
//static struct vdev fg;

//static unsigned int vbuffersize;
static int pic;
static USER_SETTINGS_T user_settings;

static void onClose(
    void);
static void quit_failure(
    void);
static void onIdle(
    void);
static void onReshape(
    int w,
    int h);
static void export(
    void);
static void onDisplay(
    void);
static int vdev_init(
    void);
static int vdev_open(
    void);
static int vdev_close(
    void);
static void next_snimek(
    void);
static void grab(
    void);

/*
static void pamet_alloc (int velikost);
static void pamet_free (void);
*/
static void data_open(
    void);
static void data_close(
    void);

/* Deprecate API V4L1 */
//static int vdev_print_info (struct vdev *fg);

/**
  * Uklizeni pri ukonceni programu.
 */
static void
onClose(
    void)
{
    remove_pid();
    ipc_remove();
    data_close();

    /*
     * if (impo.run) {
     * pamet_free ();
     * }
     */

    exit(EXIT_SUCCESS);
}

/**
  *
 */
static void
quit_failure(
    void)
{
    system("xterm -e './opso --log' &");
    onClose();
}

/**
  *
 */
static void
onIdle(
    void)
{
    glutPostRedisplay();
}

/**
  *
 */
static void
onReshape(
    int w,
    int h)
{
    window.width = w;
    window.height = h;

    glViewport(0, 0, w, h);     /* viditelna oblast pres cele okno                    */
    glMatrixMode(GL_PROJECTION);        /* zacatek modifikace projekcni matice                */
    glLoadIdentity();           /* vymazani projekcni matice (=identita)              */
    glScalef(1, -1, 1);         /* inverze y-ove osy, aby se y zvetsovalo smerem dolu */
    glOrtho(0, w, 0, h, -1, 1); /* mapovani abstraktnich souradnic do souradnic okna  */
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();           /* vymazani projekcni matice (=identita)              */
}

/**
  *
 */
static void
export(
    void)
{
    struct tm *p_tm;
    time_t cas;
    int velikost;

    if ((pix.save_TB) && (expo.p_fw == NULL)) {
        (void) time(&cas);
        p_tm = gmtime(&cas);

        if (getenv("OPSO_OUTPUT_PATH") == NULL) {
            setenv("OPSO_OUTPUT_PATH", "/tmp", 1);
        }

        velikost = strlen(getenv("OPSO_OUTPUT_PATH")) + 26 + 1;

        if ((expo.p_soubor = (char *) malloc(velikost)) == NULL) {
            syslog(LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
            quit_failure();
        }

        sprintf(expo.p_soubor, "%s/%02d_%02d_%02d_%02d_%02d_%02d.opso",
                getenv("OPSO_OUTPUT_PATH"), p_tm->tm_year + 1900,
                p_tm->tm_mon + 1, p_tm->tm_mday, p_tm->tm_hour, p_tm->tm_min,
                p_tm->tm_sec);

        if ((expo.p_fw = fopen(expo.p_soubor, "w")) == NULL) {
            syslog(LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
            quit_failure();
        }

        syslog(LOG_INFO, "EXPORT do %s start", expo.p_soubor);
    }
    else if ((!pix.save_TB) && (expo.p_fw != NULL)) {
        if (fclose(expo.p_fw) == EOF) {
            syslog(LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
        }

        syslog(LOG_INFO, "EXPORT do %s stop", expo.p_soubor);
        free(expo.p_soubor);

        expo.p_soubor = NULL;
        expo.p_fw = NULL;
    }
}

/**
  *
 */
static void
onDisplay(
    void)
{
    int snimek;

    lock(pix_semid, 0);
    if ((p_pix_sm->input_RB != pix.input_RB)
        || strcmp(p_pix_sm->str_soubor_in, pix.str_soubor_in)) {
        data_close();
        if (p_pix_sm->input_RB == 2) {
            impo.run = 1;
        }
#ifndef DBG
        else if (p_pix_sm->input_RB == 1) {
            strcpy(dev_video, "/dev/video1");
            impo.run = 0;
        }
        else {
            strcpy(dev_video, "/dev/video0");
            impo.run = 0;
        }
#endif
        pix = *p_pix_sm;
        data_open();
    }
    else {
        pix = *p_pix_sm;
    }
    unlock(pix_semid, 0);

    lock(ccbuf_semid, 0);
    snimek = p_ccbuf_sm->snimek;
    unlock(ccbuf_semid, 0);

    export();
    grab();

    glClear(GL_COLOR_BUFFER_BIT);
    glPushMatrix();

    glRasterPos2i(0, window.height);
    glPixelZoom((float) window.width / VWIDTH,
                (float) window.height / VHEIGHT);
    glDrawPixels(VWIDTH, VHEIGHT, GL_LUMINANCE, GL_UNSIGNED_BYTE,
                 p_ccbuf_sm->pole + snimek);

    next_snimek();

    /* Blue zoom cursor
     *
     * +-----------+
     * |           |
     * |     |     |
     * |   --+--   |
     * |     |     |
     * |           |
     * +-----------+
     */
    if (pix.show_cursor_zoom_CB) {
        glColor3f(0.0, 0.0, 1.0);
        glScalef((float) window.width / VWIDTH,
                 (float) window.height / VHEIGHT, 0.0);
        glBegin(GL_LINE_LOOP);
        glVertex2i(pix.zoom_x_SL - pix.zoom_size_SL / 2,
                   pix.zoom_y_SL - pix.zoom_size_SL / 2);
        glVertex2i(pix.zoom_x_SL + pix.zoom_size_SL / 2,
                   pix.zoom_y_SL - pix.zoom_size_SL / 2);
        glVertex2i(pix.zoom_x_SL + pix.zoom_size_SL / 2,
                   pix.zoom_y_SL + pix.zoom_size_SL / 2);
        glVertex2i(pix.zoom_x_SL - pix.zoom_size_SL / 2,
                   pix.zoom_y_SL + pix.zoom_size_SL / 2);
        glEnd();

        glColor3f(0.0, 0.0, 1.0);
        glBegin(GL_LINES);
        glVertex2i(pix.zoom_x_SL, pix.zoom_y_SL - 20);
        glVertex2i(pix.zoom_x_SL, pix.zoom_y_SL + 20);
        glVertex2i(pix.zoom_x_SL - 20, pix.zoom_y_SL);
        glVertex2i(pix.zoom_x_SL + 20, pix.zoom_y_SL);
        glEnd();
    }

    /* Red target
     *
     *   |
     * --+--
     *   |
     */
    if (pix.target_CB) {
        glColor3f(1.0, 0.0, 0.0);
        glBegin(GL_LINES);
        glVertex2i(pix.target_x_SL, pix.target_y_SL - 40);
        glVertex2i(pix.target_x_SL, pix.target_y_SL + 40);
        glVertex2i(pix.target_x_SL - 40, pix.target_y_SL);
        glVertex2i(pix.target_x_SL + 40, pix.target_y_SL);
        glEnd();
    }

    glPopMatrix();
    glutSwapBuffers();

    if (user_settings.usleep > 0) {
        usleep(user_settings.usleep);
    }
#ifdef DBG
    //usleep (25000);
    sleep(1);
#endif
}

static int
xioctl(
    int fd,
    int request,
    void *arg)
{
    int r;

    do
        r = ioctl(fd, request, arg);
    while (r == -1 && errno == EINTR);

    return r;
}

static int
init_read(
    unsigned int buffer_size)
{
    buffers = calloc(1, sizeof(*buffers));

    syslog(LOG_INFO, "buffer_size = %i", buffer_size);

    if (!buffers) {
        syslog(LOG_ERR, "ERROR in %s line %i (Out of memory, %m)", __FILE__,
               __LINE__);
        return -1;
    }

    buffers[0].length = buffer_size;
    buffers[0].start = malloc(buffer_size);

    if (!buffers[0].start) {
        syslog(LOG_ERR, "ERROR in %s line %i (Out of memory, %m)", __FILE__,
               __LINE__);
        return -1;
    }

    return 0;
}

static int
init_mmap(
    void)
{
    struct v4l2_requestbuffers req;

    CLEAR(req);

    req.count = 4;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_MMAP;

    if (xioctl(fd, VIDIOC_REQBUFS, &req) == -1) {
        if (errno == EINVAL) {
            syslog(LOG_ERR,
                   "ERROR in %s line %i (%s does not support memory mapping i/o, %m)",
                   __FILE__, __LINE__, dev_video);
            return -1;
        }
        else {
            syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_REQBUFS, %m)",
                   __FILE__, __LINE__);
            return -1;
        }
    }

    if (req.count < 2) {
        syslog(LOG_ERR,
               "ERROR in %s line %i (Insufficient buffer memory on %s, %m)",
               __FILE__, __LINE__, dev_video);
        return -1;
    }

    buffers = calloc(req.count, sizeof(*buffers));

    if (!buffers) {
        syslog(LOG_ERR, "ERROR in %s line %i (Out of memory, %m)", __FILE__,
               __LINE__);
        return -1;
    }

    for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
        struct v4l2_buffer buf;

        CLEAR(buf);

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = n_buffers;

        if (xioctl(fd, VIDIOC_QUERYBUF, &buf) == -1) {
            syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_QUERYBUF, %m)",
                   __FILE__, __LINE__);
            return -1;
        }

        buffers[n_buffers].length = buf.length;
        buffers[n_buffers].start = mmap(NULL /* start anywhere */ ,
                                        buf.length,
                                        PROT_READ | PROT_WRITE /* required */
                                        ,
                                        MAP_SHARED /* recommended */ ,
                                        fd, buf.m.offset);

        if (buffers[n_buffers].start == MAP_FAILED) {
            syslog(LOG_ERR, "ERROR in %s line %i (mmap, %m)", __FILE__,
                   __LINE__);
            return -1;
        }
    }

    return 0;
}

static int
init_userp(
    unsigned int buffer_size)
{
    struct v4l2_requestbuffers req;
    unsigned int page_size;

    page_size = getpagesize();
    buffer_size = (buffer_size + page_size - 1) & ~(page_size - 1);

    CLEAR(req);

    req.count = 4;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_USERPTR;

    if (xioctl(fd, VIDIOC_REQBUFS, &req) == -1) {
        if (errno == EINVAL) {
            syslog(LOG_ERR,
                   "ERROR in %s line %i (%s does not support user pointer i/o, %m)",
                   __FILE__, __LINE__, dev_video);
            return -1;
        }
        else {
            syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_REQBUFS, %m)",
                   __FILE__, __LINE__);
            return -1;
        }
    }

    buffers = calloc(4, sizeof(*buffers));

    if (!buffers) {
        syslog(LOG_ERR, "ERROR in %s line %i (Out of memory, %m)", __FILE__,
               __LINE__);
        return -1;
    }

    for (n_buffers = 0; n_buffers < 4; ++n_buffers) {
        buffers[n_buffers].length = buffer_size;
        buffers[n_buffers].start = memalign( /* boundary */ page_size,
                                            buffer_size);

        if (!buffers[n_buffers].start) {
            syslog(LOG_ERR, "ERROR in %s line %i (Out of memory, %m)",
                   __FILE__, __LINE__);
            return -1;
        }
    }

    return 0;
}

/* Deprecate API V4L1 */
//static int vdev_print_info (struct vdev *fg)
//{
//  struct video_capability video_info;
//  char BooleanText[2][4]={ "ANO\0",
//                           "NE \0" };
//                           
//  if (ioctl (fg->video_dev, VIDIOCGCAP, &video_info) == -1) {
//    syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
//    return -1;
//  }
//  else {
//    syslog (LOG_INFO, "Device Name: %s\n", video_info.name);
//    syslog (LOG_INFO, "Can capture: %s \n", BooleanText[ ((video_info.type & VID_TYPE_CAPTURE) == 0)]);
//    syslog (LOG_INFO, "Width min-max: %i-%i\n", video_info.minwidth, video_info.maxwidth);
//    syslog (LOG_INFO, "Height min-max: %i-%i\n", video_info.minheight, video_info.maxheight);
//  }
//
//  return 0;
//}

static int
vdev_getinfo(
    )
{
    int i;
    struct v4l2_standard standard;
    struct v4l2_input input;

    syslog(LOG_INFO, "Supported norms:");

    for (i = 0;; ++i) {
        memset(&standard, 0, sizeof(standard));
        standard.index = i;

        if (xioctl(fd, VIDIOC_ENUMSTD, &standard) == -1) {
            break;
        }
        else {
            syslog(LOG_INFO,
                   "    %s frame period of %i / %i seconds, framelines = %i",
                   standard.name, standard.frameperiod.numerator,
                   standard.frameperiod.denominator, standard.framelines);
        }
    }

    syslog(LOG_INFO, "Supported inputs:");

    for (i = 0;; ++i) {
        memset(&input, 0, sizeof(input));
        input.index = i;

        if (xioctl(fd, VIDIOC_ENUMINPUT, &input) < 0) {
            break;
        }
        else {
            syslog(LOG_INFO, "    %s = %i", input.name, input.index);
        }
    }

    return 0;
}

/**
  *
 */
static int
vdev_init(
    void)
{
    struct v4l2_capability cap;
    //struct v4l2_cropcap cropcap;
    //struct v4l2_crop crop;
    struct v4l2_format fmt;
    unsigned int min;
    //int overlay_on;
    int input;
    v4l2_std_id std_id;

    if (xioctl(fd, VIDIOC_QUERYCAP, &cap) == -1) {
        if (errno == EINVAL) {
            syslog(LOG_ERR, "ERROR in %s line %i (%s is no V4L2 device, %m)",
                   __FILE__, __LINE__, dev_video);
            return -1;
        }
        else {
            syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_QUERYCAP, %m)",
                   __FILE__, __LINE__);
            return -1;
        }
    }

    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
        syslog(LOG_ERR,
               "ERROR in %s line %i (%s is no video capture device, %m)",
               __FILE__, __LINE__, dev_video);
        return -1;
    }

    //if (!(cap.capabilities & V4L2_CAP_VIDEO_OVERLAY)) {
    //  syslog (LOG_ERR, "ERROR in %s line %i (%s is no video overlay device, %m)", __FILE__, __LINE__, dev_video);
    //  return -1;
    //}

    switch (user_settings.io_method) {
    case IO_METHOD_READ:
        if (!(cap.capabilities & V4L2_CAP_READWRITE)) {
            syslog(LOG_ERR,
                   "ERROR in %s line %i (%s does not support read i/o, %m)",
                   __FILE__, __LINE__, dev_video);
            return -1;
        }
        break;

    case IO_METHOD_MMAP:
    case IO_METHOD_USERPTR:
        if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
            syslog(LOG_ERR,
                   "ERROR in %s line %i (%s does not support streaming i/o, %m)",
                   __FILE__, __LINE__, dev_video);
            return -1;
        }
        break;
    }


    /* Select video input, video standard and tune here. */

    input = user_settings.v4l_input;
    if (xioctl(fd, VIDIOC_S_INPUT, &input) == -1) {
        syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_S_INPUT, %m)", __FILE__,
               __LINE__);
        return -1;
    }
    syslog(LOG_ERR, "VIDIOC_S_INPUT %i", user_settings.v4l_input);

    std_id = user_settings.v4l_standard;
    if (xioctl(fd, VIDIOC_S_STD, &std_id) == -1) {
        syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_S_STD, %m)", __FILE__,
               __LINE__);
        return -1;
    }
    syslog(LOG_ERR, "VIDIOC_S_STD %i", user_settings.v4l_standard);

    vdev_getinfo();

    //CLEAR (cropcap);
    //
    //cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    //
    //if (xioctl (fd, VIDIOC_CROPCAP, &cropcap) == 0) {
    //  crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    //  crop.c = cropcap.defrect; /* reset to default */
    //
    //  if (xioctl (fd, VIDIOC_S_CROP, &crop) == -1) {
    //    switch (errno) {
    //      case EINVAL:
    //        syslog (LOG_ERR, "ERROR in %s line %i (Cropping no supported, %m)", __FILE__, __LINE__);
    //        break;

    //      default:
    //        syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
    //        break;
    //    }
    //  }
    //} else {    
    //  syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
    //}

    CLEAR(fmt);

    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width = VWIDTH;
    fmt.fmt.pix.height = VHEIGHT;
    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_GREY;
    //fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;
    fmt.fmt.pix.field = V4L2_FIELD_ANY;

    if (xioctl(fd, VIDIOC_S_FMT, &fmt) == -1) {
        syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_S_FMT, %m)", __FILE__,
               __LINE__);
    }

    /* Note VIDIOC_S_FMT may change width and height. */

    //printf("width = %i\n", fmt.fmt.pix.width);
    //printf("height = %i\n", fmt.fmt.pix.height);
    //printf("sizeimage = %i\n", fmt.fmt.pix.sizeimage);
    //printf("bytesperline = %i\n", fmt.fmt.pix.bytesperline);
    //printf("field = %i\n", fmt.fmt.pix.field);

    /* Buggy driver paranoia. */
    //min = fmt.fmt.pix.width * 2;
    min = fmt.fmt.pix.width;
    if (fmt.fmt.pix.bytesperline < min)
        fmt.fmt.pix.bytesperline = min;
    min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
    if (fmt.fmt.pix.sizeimage < min)
        fmt.fmt.pix.sizeimage = min;

    switch (user_settings.io_method) {
    case IO_METHOD_READ:
        init_read(VWIDTH * VHEIGHT);
        break;

    case IO_METHOD_MMAP:
        init_mmap();
        break;

    case IO_METHOD_USERPTR:
        init_userp(fmt.fmt.pix.sizeimage);
        break;
    }

    //overlay_on = 1;
    //if (xioctl (fd, VIDIOC_OVERLAY, &overlay_on) == -1) {
    //  syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
    //  return -1;
    //}

    return 0;

    /* Deprecate API V4L1 */
    //fg.video_chan.channel = CHANNEL_COMPOSITE; 
    //fg.video_chan.norm = 3;
    //
    //if (ioctl (fg.video_dev, VIDIOCSCHAN, &fg.video_chan) == -1) {
    //  syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
    //}
}

/**
  *
 */
static int
vdev_open(
    void)
{
    struct stat st;

    if (stat(dev_video, &st) == -1) {
        syslog(LOG_ERR, "ERROR in %s line %i (Cannot identify %s, %m)",
               __FILE__, __LINE__, dev_video);
        return -1;
    }

    if (!S_ISCHR(st.st_mode)) {
        syslog(LOG_ERR, "ERROR in %s line %i (%s is no device, %m)", __FILE__,
               __LINE__, dev_video);
        return -1;
    }

    if ((fd = open(dev_video, O_RDWR | O_NONBLOCK, 0)) == -1) {
        syslog(LOG_ERR, "ERROR in %s line %i (Cannot open %s, %m)", __FILE__,
               __LINE__, dev_video);
        return -1;
    }

    syslog(LOG_INFO, "%s opened", dev_video);
    return 0;

    /* Deprecate API V4L1 */
    //if ((fg.video_dev = open (dev_video, O_RDWR)) < 0) {
    //  syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
    //  quit_failure ();
    //}
    //
    //syslog (LOG_INFO, "%s opened", dev_video);  
    //
    //if (vdev_print_info (&fg) != 0) {
    //  quit_failure ();
    //}
    //  
    //vdev_init ();
    //
    //fg.vid_mmap[0].frame = 0;
    //fg.vid_mmap[1].frame = 1;
    //fg.vid_mmap[0].width = fg.vid_mmap[1].width = VWIDTH;
    //fg.vid_mmap[0].height = fg.vid_mmap[1].height = VHEIGHT;
    //fg.vid_mmap[0].format = fg.vid_mmap[1].format = VIDEO_PALETTE_GREY;
    //
    //vbuffersize = VDEPTH * VWIDTH * VHEIGHT;
    //
    //fg.vid_mbuf.size = vbuffersize * 2;
    //fg.vid_mbuf.frames = 0;
    //fg.vid_mbuf.offsets[0] = 0;
    //fg.vid_mbuf.offsets[1] = vbuffersize;
    //
    //if (ioctl (fg.video_dev, VIDIOCGMBUF, &fg.vid_mbuf) < 0) {
    //  syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
    //}
    //
    //fg.p_video_map = mmap (0, fg.vid_mbuf.size, PROT_READ|PROT_WRITE, MAP_SHARED, fg.video_dev, 0);
    //pic = 0;
    //
    //if (ioctl (fg.video_dev, VIDIOCMCAPTURE, &fg.vid_mmap[pic]) < 0) {
    //  syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
    //}
}

static int
vdev_start_capturing(
    void)
{
    unsigned int i;
    enum v4l2_buf_type type;

    switch (user_settings.io_method) {
    case IO_METHOD_READ:
        /* Nothing to do. */
        break;

    case IO_METHOD_MMAP:
        for (i = 0; i < n_buffers; ++i) {
            struct v4l2_buffer buf;

            CLEAR(buf);

            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_MMAP;
            buf.index = i;

            if (xioctl(fd, VIDIOC_QBUF, &buf) == -1) {
                syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_QBUF, %m)",
                       __FILE__, __LINE__);
                return -1;
            }
        }

        type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (xioctl(fd, VIDIOC_STREAMON, &type) == -1) {
            syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_STREAMON, %m)",
                   __FILE__, __LINE__);
            return -1;
        }

        break;

    case IO_METHOD_USERPTR:
        for (i = 0; i < n_buffers; ++i) {
            struct v4l2_buffer buf;

            CLEAR(buf);

            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_USERPTR;
            buf.index = i;
            buf.m.userptr = (unsigned long) buffers[i].start;
            buf.length = buffers[i].length;

            if (xioctl(fd, VIDIOC_QBUF, &buf) == -1) {
                syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_QBUF, %m)",
                       __FILE__, __LINE__);
                return -1;
            }
        }

        type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (xioctl(fd, VIDIOC_STREAMON, &type) == -1) {
            syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_STREAMON, %m)",
                   __FILE__, __LINE__);
            return -1;
        }

        break;
    }

    return 0;
}

static int
vdev_stop_capturing(
    void)
{
    enum v4l2_buf_type type;

    switch (user_settings.io_method) {
    case IO_METHOD_READ:
        /* Nothing to do. */
        break;

    case IO_METHOD_MMAP:
    case IO_METHOD_USERPTR:
        type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (-1 == xioctl(fd, VIDIOC_STREAMOFF, &type)) {
            syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_STREAMOFF, %m)",
                   __FILE__, __LINE__);
            return -1;
        }

        break;
    }

    return 0;
}

static int
vdev_uninit_device(
    void)
{
    unsigned int i;

    switch (user_settings.io_method) {
    case IO_METHOD_READ:
        free(buffers[0].start);
        break;

    case IO_METHOD_MMAP:
        for (i = 0; i < n_buffers; ++i) {
            if (munmap(buffers[i].start, buffers[i].length) == -1) {
                syslog(LOG_ERR, "ERROR in %s line %i (munmap, %m)", __FILE__,
                       __LINE__);
                return -1;
            }
        }
        break;

    case IO_METHOD_USERPTR:
        for (i = 0; i < n_buffers; ++i) {
            free(buffers[i].start);
        }
        break;
    }

    free(buffers);
    return 0;
}

/**
  *
 */
static int
vdev_close(
    void)
{
    if (close(fd) == -1) {
        syslog(LOG_ERR, "ERROR in %s line %i (Close %s failure, %m)",
               __FILE__, __LINE__, dev_video);
        return -1;
    }

    fd = -1;

    syslog(LOG_INFO, "Videodevice closed");
    return 0;

    /* Deprecate API V4L1 */
    //close (fg.video_dev);
}

/**
  *
 */
static void
next_snimek(
    void)
{
    lock(ccbuf_semid, 0);
    if (p_ccbuf_sm->snimek < (SNIMKU - 1)) {
        p_ccbuf_sm->snimek++;
    }
    else {
        p_ccbuf_sm->snimek = 0;
    }
    unlock(ccbuf_semid, 0);
}

static int
wait_for_data(
    void)
{
    //unsigned int count = 100;

    while (1) {
        fd_set fds;
        struct timeval tv;
        int r;

        FD_ZERO(&fds);
        FD_SET(fd, &fds);

        /* Timeout. */
        tv.tv_sec = 2;
        tv.tv_usec = 0;

        r = select(fd + 1, &fds, NULL, NULL, &tv);

        if (r == -1) {
            if (EINTR == errno) {
                continue;
            }

            syslog(LOG_ERR, "ERROR in %s line %i (select, %m)", __FILE__,
                   __LINE__);
            return -1;
        }

        if (r == 0) {
            syslog(LOG_ERR, "ERROR in %s line %i (select timeout, %m)",
                   __FILE__, __LINE__);
            return -1;
        }

        break;
        /* EAGAIN - continue select loop. */
    }

    return 0;
}

static void
process_image(
    const void *p_image)
{
    int x;
    int y;
    int index = 0;
    unsigned char *p = (unsigned char *) p_image;

    for (y = 0; y < VHEIGHT; y++) {
        for (x = 0; x < VWIDTH; x++) {
            if (x == 0) {
                index = VWIDTH * VHEIGHT - VWIDTH * (y + 1);
            }

            p_ccbuf_sm->pole[p_ccbuf_sm->snimek][index] = *p;
            p++;

            if (pix.save_TB) {
                fwrite(p, sizeof(unsigned char), 1, expo.p_fw);
            }

            ++index;
        }
    }
}

static int
read_frame(
    void)
{
    struct v4l2_buffer buf;
    unsigned int i;

    switch (user_settings.io_method) {
    case IO_METHOD_READ:
        if (read(fd, buffers[0].start, buffers[0].length) == -1) {
            switch (errno) {
            case EAGAIN:
                return 0;

            case EIO:
                /* Could ignore EIO, see spec. */

                /* fall through */

            default:
                syslog(LOG_ERR, "ERROR in %s line %i (read, %m)", __FILE__,
                       __LINE__);
                return -1;
            }
        }

        process_image(buffers[0].start);
        break;

    case IO_METHOD_MMAP:
        CLEAR(buf);

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;

        if (-1 == xioctl(fd, VIDIOC_DQBUF, &buf)) {
            switch (errno) {
            case EAGAIN:
                return 0;

            case EIO:
                /* Could ignore EIO, see spec. */

                /* fall through */

            default:
                syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_DQBUF, %m)",
                       __FILE__, __LINE__);
                return -1;
            }
        }

        assert(buf.index < n_buffers);

        process_image(buffers[buf.index].start);

        if (xioctl(fd, VIDIOC_QBUF, &buf) == -1) {
            syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_QBUF, %m)", __FILE__,
                   __LINE__);
            return -1;
        }

        break;

    case IO_METHOD_USERPTR:
        CLEAR(buf);

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_USERPTR;

        if (xioctl(fd, VIDIOC_DQBUF, &buf) == -1) {
            switch (errno) {
            case EAGAIN:
                return 0;

            case EIO:
                /* Could ignore EIO, see spec. */

                /* fall through */

            default:
                syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_DQBUF, %m)",
                       __FILE__, __LINE__);
                return -1;
            }
        }

        for (i = 0; i < n_buffers; ++i) {
            if (buf.m.userptr == (unsigned long) buffers[i].start
                && buf.length == buffers[i].length) {
                break;
            }
        }

        assert(i < n_buffers);

        process_image((void *) buf.m.userptr);

        if (-1 == xioctl(fd, VIDIOC_QBUF, &buf)) {
            syslog(LOG_ERR, "ERROR in %s line %i (VIDIOC_QBUF, %m)", __FILE__,
                   __LINE__);
            return -1;
        }

        break;
    }

    return 1;
}

/**
  *
 */
static void
grab(
    void)
{
    unsigned char *p = NULL;
    int index = 0;
    int x;
    int y;

    if (impo.run) {
        if ((p = (unsigned char *) malloc(1)) == NULL) {
            syslog(LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
            quit_failure();
        }
    }

    if (!impo.run) {
        pic = !pic;

        if (wait_for_data() == -1) {
            return;
        }

        if (read_frame() == -1) {
            return;
        }

        /* Deprecate API V4L1 */
        //if (ioctl (fg.video_dev, VIDIOCMCAPTURE, &fg.vid_mmap[pic]) < 0) {
        //  syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
        //}
        //
        ///* Wait for Full Buffer */
        //ioctl (fg.video_dev, VIDIOCSYNC, &fg.vid_mmap[!pic]);
        //p = fg.p_video_map + fg.vid_mbuf.offsets[!pic];
    }
    else if (impo.p_fr != NULL) {
        fseek(impo.p_fr, impo.seek, SEEK_SET);
        impo.seek += VWIDTH * VHEIGHT;
    }
    else {
        syslog(LOG_ERR, "ERROR in %s line %i (Neni co cist)", __FILE__,
               __LINE__);
        quit_failure();
    }

    if (impo.run) {
        for (y = 0; y < VHEIGHT; y++) {
            for (x = 0; x < VWIDTH; x++) {
                if (x == 0) {
                    index = VWIDTH * VHEIGHT - VWIDTH * (y + 1);
                }

                if (impo.run) {
                    if (fread(p, sizeof(*p), 1, impo.p_fr) == 0) {
                        impo.seek = 0;
                        fseek(impo.p_fr, impo.seek, SEEK_SET);
                        fread(p, sizeof(*p), 1, impo.p_fr);
                        /* printf ("*** Novy cyklus ***\n"); */
                    }

                    p_ccbuf_sm->pole[p_ccbuf_sm->snimek][index] = *p;
                }

                if (pix.save_TB) {
                    fwrite(p, sizeof(*p), 1, expo.p_fw);
                }

                ++index;
            }
        }

        free(p);
    }
}

/**
  *
 */

/*static void pamet_alloc (int velikost)
{
  if ((impo.p_soubor = (char *)malloc (velikost)) == NULL) {
    syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
    quit_failure ();
  }
}*/

/**
  *
 */

/*static void pamet_free (void)
{
  free (impo.p_soubor);
}*/

/**
  *
 */
static void
data_open(
    void)
{
    if (impo.run) {
        if ((impo.p_fr = fopen(pix.str_soubor_in, "r")) == NULL) {
            syslog(LOG_ERR, "ERROR in %s line %i (File %s open failure, %m)",
                   __FILE__, __LINE__, pix.str_soubor_in);
        }
        else {
            syslog(LOG_INFO, "%s opened", pix.str_soubor_in);
        }
    }
    else {
        if (vdev_open() == -1) {
            quit_failure();
        }

        if (vdev_init() == -1) {
            quit_failure();
        }

        if (vdev_start_capturing() == -1) {
            quit_failure();
        }
    }
}

/**
  * 
 */
static void
run_long_option(
    const char *option,
    const char *optarg)
{
    if (option == "log") {
        system("./log");
    }
    else if (option == "log-pattern") {
        /*
         * char *promena;
         * char var[] = "OPSO_LOG_PATTERN=";
         * 
         * if ((promena = (char *)malloc (strlen (var) + strlen (optarg) + 1)) == NULL) {
         * syslog (LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
         * exit (EXIT_FAILURE);   
         * }
         * 
         * strcpy (promena, var);
         * strcat (promena, optarg);
         * 
         * putenv (promena);
         * system ("./log");
         * 
         * free (promena);
         */

        setenv("OPSO_LOG_PATTERN", optarg, 1);
        system("./log");
    }
    else if (option == "output-path") {
        setenv("OPSO_OUTPUT_PATH", optarg, 1);
        return;
    }
    else if (option == "version") {
        version();
    }
    else if (option == "width") {
        window.width = atoi(optarg);
        return;
    }
    else if (option == "height") {
        window.height = atoi(optarg);
        return;
    }
    else if (option == "x") {
        window.x = atoi(optarg);
        return;
    }
    else if (option == "y") {
        window.y = atoi(optarg);
        return;
    }
    else if (option == "v4l-io-method") {
        user_settings.io_method = atoi(optarg);

        switch (user_settings.io_method) {
        case IO_METHOD_USERPTR:
        case IO_METHOD_MMAP:
        case IO_METHOD_READ:
            return;

        default:
            printf("For option --v4l-io-method is %i not supproted value\n",
                   user_settings.io_method);
            break;
        }
        return;
    }
    else if (option == "v4l-input") {
        user_settings.v4l_input = atoi(optarg);
        return;
    }
    else if (option == "v4l-standard") {
        user_settings.v4l_standard = atoi(optarg);
        return;
    }
    else if (option == "v4l-device") {
        strncpy(user_settings.v4l_device, optarg, V4L_DEVICE_MAX);
        return;
    }
    else if (option == "usleep") {
        user_settings.usleep = atoi(optarg);
        return;
    }
    else if (option == "test-mode") {
        user_settings.test_mode = 1;
        return;
    }

    exit(EXIT_SUCCESS);
}

/**
  *
 */
static void
data_close(
    void)
{
    if (!impo.run) {
        if (fd != -1) {
            vdev_stop_capturing();
            vdev_uninit_device();
            vdev_close();
        }
    }
    else if (impo.p_fr != NULL) {
        if (fclose(impo.p_fr) == EOF) {
            syslog(LOG_ERR, "ERROR in %s line %i (%m)", __FILE__, __LINE__);
        }
        else {
            syslog(LOG_INFO, "%s closed", pix.str_soubor_in);
        }
    }
}

/**
  * MAIN opso
 */
int
main(
    int argc,
    char *argv[])
{
    int o;
    int i;
    char retezec[100];

    sprintf(retezec, "RUN opso %s", VERSION);

#ifdef DBG
    strcat(retezec, "dbg");
#endif

    strcat(retezec, CAS_KOMPILACE);

    openlog("opso", LOG_PID | LOG_CONS, LOG_LOCAL0);
    syslog(LOG_INFO, retezec);

    window.width = VWIDTH;
    window.height = VHEIGHT;
    window.x = 0;
    window.y = 0;
    impo.run = 0;
    strcpy(dev_video, "/dev/video0");

    user_settings.test_mode = 0;
    user_settings.usleep = 0;
    user_settings.io_method = IO_METHOD_READ;
    //user_settings.io_method = IO_METHOD_MMAP;
    //user_settings.io_method = IO_METHOD_USERPTR;
    user_settings.v4l_input = 1;
    user_settings.v4l_standard = V4L2_STD_PAL_Nc;
    memset(user_settings.v4l_device, 0, V4L_DEVICE_MAX);

    while (1) {
        int option_index = 0;
        static struct option long_options[] = {
            {"help", 0, 0, 'h'},
            {"version", 0, 0, 0},
            {"test-mode", 0, 0, 0},
            {"log", 0, 0, 0},
            {"log-pattern", 1, 0, 0},
            {"input", 1, 0, 'i'},
            {"output-path", 1, 0, 0},
            {"width", 1, 0, 0},
            {"height", 1, 0, 0},
            {"x", 1, 0, 0},
            {"y", 1, 0, 0},
            {"v4l-input", 1, 0, 0},
            {"v4l-standard", 1, 0, 0},
            {"v4l-device", 1, 0, 0},
            {"v4l-io-method", 1, 0, 0},
            {"usleep", 1, 0, 0},
            {0, 0, 0, 0}
        };

        o = getopt_long(argc, argv, "hi:", long_options, &option_index);
        if (o == -1) {
            break;
        }

        switch (o) {
        case 0:
            run_long_option(long_options[option_index].name, optarg);
            break;

        case 'h':
            napoveda();
            exit(EXIT_SUCCESS);
            break;

        case 'i':
            /*
             * pamet_alloc (strlen (optarg) + 1);
             * strcpy (impo.p_soubor, optarg);
             */

            if (strlen(optarg) < MAX_LEN_FILE_NAME) {
                impo.run = 1;
                strcpy(pix.str_soubor_in, optarg);
            }

            break;

        case '?':
            break;

        default:
            syslog(LOG_INFO, "getopt vrtila 0%o", o);
            break;
        }

        /*
         * while (optind < argc) {
         * syslog (LOG_INFO, "Neplatny argument: %s", argv[optind++]);
         * }
         */
    }

    if (user_settings.v4l_device[0] != '\0') {
        strcpy(dev_video, user_settings.v4l_device);
    }

    ipc_create();
    strcpy(p_pix_sm->str_soubor_in, pix.str_soubor_in);

    p_pix_sm->brightness_SL = 0;
    p_pix_sm->contrast_SL = 255.0;
    p_pix_sm->average_SL = 25;
    p_pix_sm->zmena_pru = 0;
    p_pix_sm->zoom_x_SL = 260;
    p_pix_sm->zoom_y_SL = 260;
    p_pix_sm->zoom_size_SL = 260;
    p_pix_sm->zoom_cut_x_SL = 0;
    p_pix_sm->zoom_cut_y_SL = 0;
    p_pix_sm->zoom_cut_x_size_SL = 10;
    p_pix_sm->zoom_cut_y_size_SL = 10;
    p_pix_sm->slith_begin_SL = 319;
    p_pix_sm->slith_end_SL = 338;
    p_pix_sm->cut_x_SL = 300;
    p_pix_sm->cut_y_SL = 318;
    p_pix_sm->target_x_SL = 334;
    p_pix_sm->target_y_SL = 300;
    p_pix_sm->faulty_pix_x_SL = 10;
    p_pix_sm->faulty_pix_y_SL = 10;
    p_pix_sm->faulty_pix_size_SL = 2;
    p_pix_sm->zoom_CB = 0;
    p_pix_sm->zoom_bind_target_CB = 1;
    p_pix_sm->target_CB = 1;
    p_pix_sm->show_cut_xy_CB = 1;
    p_pix_sm->show_fit_gaus_xy_CB = 1;
    p_pix_sm->show_cursor_slith_CB = 0;
    p_pix_sm->show_cursor_zoom_CB = 1;
    p_pix_sm->show_cursor_cut_xy_CB = 0;
    p_pix_sm->show_cube_position_CB = 0;
    p_pix_sm->show_value_in_quadrant_CB = 0;
    p_pix_sm->show_faulty_pix_CB = 0;
    p_pix_sm->show_cursor_faulty_pix_CB = 0;
    p_pix_sm->histogram_CB = 0;
    p_pix_sm->sliding_average_RB = 1;
    p_pix_sm->input_RB = 0;
    p_pix_sm->save_TB = 0;
    p_pix_sm->timeout_tsgc_SB = 3;
    p_pix_sm->count_step_SB = 1;
    p_pix_sm->max_count_step_SB = 50;
    p_pix_sm->permissible_deviation_SB = 0;
    p_pix_sm->sec_in_pix_SB = 24;
    p_pix_sm->min_intenzity_star_SB = 0;
    p_pix_sm->max_fwhm_star_SB = 300;
    p_pix_sm->corr_at_error_SB = 0;
    p_pix_sm->allowed_offset_SB = 6;
    p_pix_sm->timeout_allowed_offset_SB = 5;
    p_pix_sm->calibration = 0;

    for (i = 0; i < FAULTY_PIX_MAX; ++i) {
        p_pix_sm->pole_faulty_pix[i][0] = 0;
        p_pix_sm->pole_faulty_pix[i][1] = 0;
        p_pix_sm->pole_faulty_pix[i][2] = 0;
    }

    //strcpy (p_pix_sm->str_soubor_in, "/dev/video0");
    strcpy(p_pix_sm->str_soubor_out, "");

    impo.p_fr = NULL;
    impo.seek = 0;
    expo.p_fw = NULL;
    p_ccbuf_sm->snimek = 0;

    lock(pix_semid, 0);
    pix = *p_pix_sm;
    unlock(pix_semid, 0);

    data_open();
    write_pid("opso", getpid());

    if (!user_settings.test_mode) {
        system("./pointace.sh &");
    }

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
    glutInitWindowSize(window.width, window.height);
    glutInitWindowPosition(window.x, window.y);
    sprintf(retezec, "OpenGL Pointing System for Ondrejov %s", VERSION);

#ifdef DBG
    strcat(retezec, "dbg");
#endif

    glutCreateWindow(retezec);
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glShadeModel(GL_FLAT);
    glutDisplayFunc(onDisplay);
    glutReshapeFunc(onReshape);
    glutIdleFunc(onIdle);
    glutCloseFunc(onClose);
    glutMainLoop();

    return 0;
}
