
/**
  * Author: Jan Fuchs
  * $Date: 2009-04-08 17:35:05 +0200 (Wed, 08 Apr 2009) $
  * $Rev: 12 $
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <pthread.h>
#include <semaphore.h>
#include <assert.h>
#include <string.h>
#include <GL/freeglut.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include <sys/sem.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <getopt.h>

#include "../fitgsl/fitgsl.h"
#include "global.h"
#include "ipc.h"
#include "tty.h"

#define W_MEZERA         0
#define W_REZ            255
#define COMMAND_MAX      1023
#define TSGC_HISTORY_LEN 10

#define AKTUALNI_SNIMEK_SM ((unsigned char *)(p_ccbuf_sm->pole + snimek))

static struct {
    int width;
    int height;
    int x;
    int y;
} window;

static struct {
    int snimek;
    int index;
    int index_old;
    int naplnen;
    int historie[SNIMKU];
    int prumer[VWIDTH * VHEIGHT];
    unsigned char pole[VWIDTH * VHEIGHT];
    unsigned char zoom[VWIDTH * VHEIGHT];
    unsigned char *p_pole;
} buffer;

static struct {
    int x;
    int y;
    int max;
} kostka;

static struct {
    int a;
    int b;
    int c;
    int d;
} kvadrant;

#define W_WIDTH  (VWIDTH + W_MEZERA + W_REZ)
#define W_HEIGHT (VHEIGHT + W_MEZERA + W_REZ)
#define VWIDTH_ZOOM  (window.width - W_MEZERA - (W_REZ * ((float)window.width / W_WIDTH)))
#define VHEIGHT_ZOOM (window.height - W_MEZERA - (W_REZ * ((float)window.height / W_HEIGHT)))

PIX pix;
CCBUF ccbuf;
static int snimek = 0;
static int intenzity_star = 0;
static float fwhm = 0;
static char tsgc_history[TSGC_HISTORY_LEN][COMMAND_MAX + 1];
static char command[COMMAND_MAX + 1];
static fitgsl_data *fitgaus_data_x = NULL;
static fitgsl_data *fitgaus_data_y = NULL;

void *shared_memory = (void *) 0;

int compare_function(
    const void *a,
    const void *b);

static void konec(
    int sig);
static void onClose(
    void);
static void onIdle(
    void);
static void onReshape(
    int w,
    int h);
static void zoom(
    unsigned char *p_pole,
    unsigned char *p_out,
    int zoom_size);
static void median(
    void);
static void prumerovani(
    void);
//static void automation_pointing (int *p_pozice, int start, int posun, int port);
static void onDisplay(
    void);
static unsigned char *pointer(
    void);
static int rozsah(
    int cislo_x,
    int cislo_y,
    int max_x,
    int max_y);

/**
  *
 */
static void
konec(
    int sig)
{
    //tty_write (0, 2, 0);
    //tty_write (1, 2, 0);
    //tty_write (0, 1, pix.kostka_start_x);
    //tty_write (1, 1, pix.kostka_start_y);
    //tty_close ();
    //fitgsl_free_data (fitgaus_data_x);
    //fitgsl_free_data (fitgaus_data_y);
    exit(EXIT_SUCCESS);
}

/**
  *
 */
static void
onClose(
    void)
{
    konec(SIGTERM);
}

/**
  *
 */
static void
onIdle(
    void)
{
    glutPostRedisplay();
}

/**
  *
 */
static void
onReshape(
    int w,
    int h)
{
    window.width = w;
    window.height = h;

    glViewport(0, 0, w, h);     /* viditelna oblast pres cele okno                    */
    glMatrixMode(GL_PROJECTION);        /* zacatek modifikace projekcni matice                */
    glLoadIdentity();           /* vymazani projekcni matice (=identita)              */
    glScalef(1, -1, 1);         /* inverze y-ove osy, aby se y zvetsovalo smerem dolu */
    glOrtho(0, w, 0, h, -1, 1); /* mapovani abstraktnich souradnic do souradnic okna  */
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();           /* vymazani projekcni matice (=identita)              */
}

/**
  *  Vyrizne ze snimku cast, ktera se ma zoomovat a ulozi ji do buffer.zoom[]
 */
static void
zoom(
    unsigned char *p_pole,
    unsigned char *p_out,
    int zoom_size)
{
    int x;
    int y;
    int x_zoom = 0;
    int y_zoom = 0;
    int a_poc = 0;
    int b_poc = 0;
    int c_poc = 0;
    int d_poc = 0;

    kvadrant.a = 0;
    kvadrant.b = 0;
    kvadrant.c = 0;
    kvadrant.d = 0;

    pix.zoom_y_SL = VHEIGHT - pix.zoom_y_SL;
    for (y = 0; y < VWIDTH; y++) {
        for (x = 0; x < VHEIGHT; x++) {
            if ((x > (pix.zoom_y_SL - zoom_size / 2)) && (x <= pix.zoom_y_SL)
                && (y >= pix.zoom_x_SL - zoom_size / 2)
                && (y < pix.zoom_x_SL)) {
                kvadrant.c += p_pole[x * VWIDTH + y];
                ++c_poc;
            }
            else if ((x > pix.zoom_y_SL)
                     && (x <= (pix.zoom_y_SL + zoom_size / 2))
                     && (y >= pix.zoom_x_SL - zoom_size / 2)
                     && (y < pix.zoom_x_SL)) {
                kvadrant.a += p_pole[x * VWIDTH + y];
                ++a_poc;
            }
            else if ((x > (pix.zoom_y_SL - zoom_size / 2))
                     && (x <= pix.zoom_y_SL) && (y >= pix.zoom_x_SL)
                     && (y < (pix.zoom_x_SL + zoom_size / 2))) {
                kvadrant.d += p_pole[x * VWIDTH + y];
                ++d_poc;
            }
            else if ((x > pix.zoom_y_SL)
                     && (x <= (pix.zoom_y_SL + zoom_size / 2))
                     && (y >= pix.zoom_x_SL)
                     && (y < (pix.zoom_x_SL + zoom_size / 2))) {
                kvadrant.b += p_pole[x * VWIDTH + y];
                ++b_poc;
            }


            if ((x > (pix.zoom_y_SL - zoom_size / 2))
                && (x <= (pix.zoom_y_SL + zoom_size / 2))
                && (y >= pix.zoom_x_SL - zoom_size / 2)
                && (y < (pix.zoom_x_SL + zoom_size / 2))) {
                p_out[x_zoom * zoom_size + y_zoom] = p_pole[x * VWIDTH + y];
                if (x_zoom < (zoom_size - 1)) {
                    ++x_zoom;
                }
                else {
                    x_zoom = 0;
                    ++y_zoom;
                }
            }
        }
    }

    kvadrant.a /= a_poc;
    kvadrant.b /= b_poc;
    kvadrant.c /= c_poc;
    kvadrant.d /= d_poc;

    if (pix.automation_pointing_TB) {
        //automation_pointing (&kostka.x, pix.kostka_start_x, ((kvadrant.a + kvadrant.c) - (kvadrant.b + kvadrant.d)) * -1, 0);
        //automation_pointing (&kostka.y, pix.kostka_start_y, (kvadrant.a + kvadrant.b) - (kvadrant.c + kvadrant.d), 1);
    }
}

/**
  *
 */
int
compare_function(
    const void *a,
    const void *b)
{
    unsigned char *x = (unsigned char *) a;
    unsigned char *y = (unsigned char *) b;

    return *x - *y;
}

/**
  *
 */
static void
median(
    void)
{
    int x;
    int y;
    int i;
    int index;
    int pocet;
    int pozice;
    unsigned char pole[SNIMKU];

    pocet = pix.average_SL;
    index = 0;
    for (x = 0; x < VWIDTH; x++) {
        for (y = 0; y < VHEIGHT; y++) {
            buffer.index = x * VHEIGHT + y;

            for (i = 0; i < pocet; i++) {
                pozice = snimek - i;

                if (pozice < 0) {
                    pozice += SNIMKU;
                }

                pole[i] = p_ccbuf_sm->pole[pozice][index];
            }

            qsort(pole, pocet, sizeof(unsigned char), compare_function);
            buffer.pole[buffer.index] = pole[pocet / 2];
            ++index;
        }
    }
}

/**
  * Do buffer.pole[] ulozi zprumerovany snimek. V buffer.prumer[] si
  * udrzuje aktualni prumer snimku. Na pocatku se buffer.prumer[] plni
  * (tj. snimky se pouze pricitaji), jakmile pocet snimku koresponduje
  * s pozadovanym prumerovanim, tak se nejstarsi snimek odecte a nejnovejsi
  * snimek ze sdilene pameti se pricte.
 */
static void
prumerovani(
    void)
{
    int x;
    int y;
    int index;
    int snimek_remove;
    int pocet;

    if (pix.zmena_pru) {
        buffer.snimek = 0;
        buffer.naplnen = 0;
    }
    else {
        if (buffer.snimek < (SNIMKU - 1)) {
            ++buffer.snimek;
        }
        else {
            buffer.snimek = 0;
        }
    }

    if ((buffer.snimek + 1) == (pix.average_SL + 1)) {
        buffer.naplnen = 1;
    }

    if (buffer.naplnen == 1) {
        pocet = pix.average_SL;
    }
    else {
        pocet = buffer.snimek + 1;
    }

    snimek_remove = buffer.snimek - pix.average_SL;
    if (snimek_remove < 0) {
        snimek_remove += SNIMKU;
    }

    index = 0;
    for (x = 0; x < VWIDTH; x++) {
        for (y = 0; y < VHEIGHT; y++) {
            buffer.index = x * VHEIGHT + y;

            if ((buffer.naplnen == 0) && (buffer.snimek == 0)) {
                buffer.prumer[buffer.index] =
                    (int) p_ccbuf_sm->pole[snimek][index];
            }
            else {
                buffer.prumer[buffer.index] +=
                    (int) p_ccbuf_sm->pole[snimek][index];
            }

            if (buffer.naplnen == 1) {
                buffer.prumer[buffer.index] -=
                    (int) ccbuf.pole[snimek_remove][index];
            }

            ccbuf.pole[buffer.snimek][index] =
                p_ccbuf_sm->pole[snimek][index];
            buffer.pole[buffer.index] =
                (unsigned char) (buffer.prumer[buffer.index] / pocet);
            ++index;
        }
    }
}

/**
  * Pomoci GUI (../python/ovladani.py) uzivatel ovlivni na jaky
  * snimek vrati tato funkce ukazatel.
  *
  * Jsou 4 moznosti - cely prumerovany
  *                 - cely neprumerovany
  *                 - zoomovany prumerovany
  *                 - zoomovany neprumerovany
 */
static unsigned char *
pointer(
    void)
{
    if (pix.average_SL == 1) {
        if (pix.zoom_CB == 1) {
            pix.zoom_size_SL -= pix.zoom_size_SL % 4;
            zoom((unsigned char *) (p_ccbuf_sm->pole + snimek), buffer.zoom,
                 pix.zoom_size_SL);
            return buffer.zoom;
        }
        else {
            return AKTUALNI_SNIMEK_SM;
        }
    }
    else {
        if (pix.sliding_average_RB) {
            prumerovani();
        }
        else {
            median();
        }

        if (pix.zoom_CB == 1) {
            pix.zoom_size_SL -= pix.zoom_size_SL % 4;
            zoom(buffer.pole, buffer.zoom, pix.zoom_size_SL);
            return buffer.zoom;
        }
        else {
            return buffer.pole;
        }
    }
}

/**
  * Zjisti jak ma posunout kostky a odesle prikaz servru
  * pro rizeni kostek (../src/python/kostky_server.py).
 */
//static void automation_pointing (int *p_pozice, int start, int posun, int port)
//{
//  int nova_pozice;
//  //int smer;
//  //unsigned char prikaz[11];
//
//  if (posun > 0) {
//    nova_pozice = *p_pozice + pix.count_step_SB;
//    //smer = kladny;
//  }
//  else {
//    nova_pozice = *p_pozice - pix.count_step_SB;
//    posun *= -1;
//    //smer = zaporny;
//  }
//  
//  if (posun > pix.permissible_deviation_SB) {
//    if ((nova_pozice > kostka.max) || (nova_pozice < -kostka.max)) {
//      lock (pix_semid, 0);
//      p_pix_sm->automation_pointing_TB = 0;
//      p_pix_sm->calibration = 1;
//      unlock (pix_semid, 0);
//      system ("aplay ../sound/automation_off.wav &");
//    }
//    else {
//      *p_pozice = nova_pozice;
//      //tty_write (port, 1, start + nova_pozice);
//    }
//  }
//}

/**
  * Zajisti, aby se nekreslilo mimo vyhrazenou oblast, urcenou
  * parametry teto funkce.
 */
static int
rozsah(
    int cislo_x,
    int cislo_y,
    int max_x,
    int max_y)
{
    if ((cislo_x < 0) ||
        (cislo_y < 0) || (cislo_x > max_x) || (cislo_y > max_y)) {
        return 0;
    }
    else {
        return 1;
    }
}

/**
  *
 */
static void
printString(
    int x,
    int y,
    char *text)
{
    glRasterPos2i(x, y);
    while (*text) {
        glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *text++);
    }
}

/**
  *
 */
static void
printInt(
    int x,
    int y,
    int cislo)
{
    char text[20];

    snprintf(text, 19, "%i", cislo);
    printString(x, y, text);
}

/**
  *
 */
static void
printFWHM(
    int x,
    int y,
    float cislo)
{
    char text[100];

    snprintf(text, 99, "FWHM = %0.1f", cislo);
    printString(x, y, text);
    snprintf(text, 99, "Intenzity star = %i", intenzity_star);
    printString(x, y + 20, text);
}

/**
  *
 */
static int
souradnice_x(
    int x,
    float pomer)
{
    if (pix.zoom_CB) {
        return (x - (pix.zoom_x_SL - pix.zoom_size_SL / 2)) * pomer;
    }
    else {
        return x;
    }
}

/**
  *
 */
static int
souradnice_y(
    int y,
    float pomer)
{
    if (pix.zoom_CB) {
        return (y - (VHEIGHT - pix.zoom_y_SL - pix.zoom_size_SL / 2)) * pomer;
    }
    else {
        return y;
    }
}

static void
draw_cursor_cut_xy(
    void)
{
    int x1 = pix.cut_x_SL + pix.cut_x_size_SL;
    int x2 = pix.cut_x_SL - pix.cut_x_size_SL;
    int y1 = pix.cut_y_SL + pix.cut_y_size_SL;
    int y2 = pix.cut_y_SL - pix.cut_y_size_SL;
    int cut_y1 = pix.cut_y_SL - (pix.cut_y_slith_size_SL / 2);
    int cut_y2 = pix.cut_y_SL + (pix.cut_y_slith_size_SL / 2);

    /* Green cursor cut XY
     *
     *  |
     *  |
     * -+------
     *  |
     *  |
     */
    if (pix.show_cursor_cut_xy_CB) {
        glColor3f(0.0, 1.0, 0.0);
        glBegin(GL_LINES);
        glVertex2i(0, pix.cut_x_SL);
        glVertex2i(VWIDTH, pix.cut_x_SL);

        glVertex2i(0, x1);
        glVertex2i(VWIDTH, x1);
        glVertex2i(0, x2);
        glVertex2i(VWIDTH, x2);

        glVertex2i(cut_y1, 0);
        glVertex2i(cut_y1, VHEIGHT);
        glVertex2i(cut_y2, 0);
        glVertex2i(cut_y2, VHEIGHT);

        glVertex2i(y1, 0);
        glVertex2i(y1, VHEIGHT);
        glVertex2i(y2, 0);
        glVertex2i(y2, VHEIGHT);
        glEnd();
    }
}

static void
draw_cursor_slith(
    void)
{
    /* Violet cursor slith XY
     *
     *  | |
     *  | |
     *  | |
     *  | |
     *  | |
     */
    if (pix.show_cursor_slith_CB) {
        glColor3f(1.0, 0.4, 1.0);
        glBegin(GL_LINES);
        glVertex2i(pix.slith_begin_SL, 0);
        glVertex2i(pix.slith_begin_SL, VHEIGHT);
        glVertex2i(pix.slith_end_SL, 0);
        glVertex2i(pix.slith_end_SL, VHEIGHT);
        glEnd();
    }
}

static void
draw_blacken_faulty_pix(
    float pomer_x,
    float pomer_y)
{
    int i;
    int size;
    int faulty_pix_x;
    int faulty_pix_y;

    /* Blacken faulty pix
     *
     * +---+
     * |   |
     * +---+
     */
    if (pix.blacken_faulty_pix_CB) {
        glColor3f(0.0, 0.0, 0.0);
        for (i = 0; i < FAULTY_PIX_MAX; ++i) {
            size = pix.pole_faulty_pix[i][2] / 2;
            faulty_pix_x = souradnice_x(pix.pole_faulty_pix[i][0], pomer_x);
            faulty_pix_y = souradnice_y(pix.pole_faulty_pix[i][1], pomer_y);

            glBegin(GL_QUADS);
            glVertex2i(faulty_pix_x - size, faulty_pix_y - size);
            glVertex2i(faulty_pix_x + size, faulty_pix_y - size);
            glVertex2i(faulty_pix_x + size, faulty_pix_y + size);
            glVertex2i(faulty_pix_x - size, faulty_pix_y + size);
            glEnd();
        }
    }
}

static void
draw_faulty_pix(
    float pomer_x,
    float pomer_y)
{
    int i;
    int size;
    int faulty_pix_x;
    int faulty_pix_y;

    /* 00FFFF faulty pix
     *
     * +---+
     * |   |
     * +---+
     */
    if (pix.show_faulty_pix_CB) {
        glColor3f(0.0, 1.0, 1.0);
        for (i = 0; i < FAULTY_PIX_MAX; ++i) {
            size = pix.pole_faulty_pix[i][2] / 2;
            faulty_pix_x = souradnice_x(pix.pole_faulty_pix[i][0], pomer_x);
            faulty_pix_y = souradnice_y(pix.pole_faulty_pix[i][1], pomer_y);
            glBegin(GL_LINE_LOOP);
            glVertex2i(faulty_pix_x - size, faulty_pix_y - size);
            glVertex2i(faulty_pix_x + size, faulty_pix_y - size);
            glVertex2i(faulty_pix_x + size, faulty_pix_y + size);
            glVertex2i(faulty_pix_x - size, faulty_pix_y + size);
            glEnd();
        }
    }
}

static void
draw_cursor_faulty_pix(
    float pomer_x,
    float pomer_y)
{
    int size;
    int faulty_pix_x;
    int faulty_pix_y;

    /* Green cursor faulty pix
     *
     * +---+
     * |   |
     * +---+
     */
    if (pix.show_cursor_faulty_pix_CB) {
        glColor3f(0.0, 1.0, 0.0);
        size = pix.faulty_pix_size_SL / 2;
        faulty_pix_x = souradnice_x(pix.faulty_pix_x_SL, pomer_x);
        faulty_pix_y = souradnice_y(pix.faulty_pix_y_SL, pomer_y);

        glBegin(GL_LINE_LOOP);
        glVertex2i(faulty_pix_x - size, faulty_pix_y - size);
        glVertex2i(faulty_pix_x + size, faulty_pix_y - size);
        glVertex2i(faulty_pix_x + size, faulty_pix_y + size);
        glVertex2i(faulty_pix_x - size, faulty_pix_y + size);
        glEnd();
    }
}

static void
draw_cube_position(
    void)
{
    float pomer;

    if (pix.show_cube_position_CB) {
        glColor3f(1.0, 1.0, 1.0);
        glBegin(GL_QUADS);
        glVertex2i(W_WIDTH - 245, W_HEIGHT - 175);
        glVertex2i(W_WIDTH - 245, W_HEIGHT - 240);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 240);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 175);
        glEnd();

        glColor3f(1.0, 1.0, 1.0);
        glBegin(GL_QUADS);
        glVertex2i(W_WIDTH - 245, W_HEIGHT - 50);
        glVertex2i(W_WIDTH - 245, W_HEIGHT - 115);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 115);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 50);
        glEnd();

        glColor3f(0.0, 0.0, 0.0);
        glBegin(GL_LINES);
        glVertex2i(W_WIDTH - 245, W_HEIGHT - 175);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 175);
        glVertex2i(W_WIDTH - 245, W_HEIGHT - 165);
        glVertex2i(W_WIDTH - 245, W_HEIGHT - 175);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 165);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 175);
        glVertex2i(W_WIDTH - 130, W_HEIGHT - 165);
        glVertex2i(W_WIDTH - 130, W_HEIGHT - 175);

        glVertex2i(W_WIDTH - 245, W_HEIGHT - 50);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 50);
        glVertex2i(W_WIDTH - 245, W_HEIGHT - 40);
        glVertex2i(W_WIDTH - 245, W_HEIGHT - 50);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 40);
        glVertex2i(W_WIDTH - 15, W_HEIGHT - 50);
        glVertex2i(W_WIDTH - 130, W_HEIGHT - 40);
        glVertex2i(W_WIDTH - 130, W_HEIGHT - 50);
        glEnd();

        glColor3f(0.0, 0.0, 0.0);
        printString(W_WIDTH - 130, W_HEIGHT - 150, "0");
        printString(W_WIDTH - 130, W_HEIGHT - 25, "0");

        glBegin(GL_LINES);
        pomer = 230.0 / (kostka.max * 2);
        glVertex2f(W_WIDTH - 130 + kostka.x * pomer, W_HEIGHT - 175);
        glVertex2f(W_WIDTH - 130 + kostka.x * pomer, W_HEIGHT - 240);
        glVertex2f(W_WIDTH - 130 + kostka.y * pomer, W_HEIGHT - 50);
        glVertex2f(W_WIDTH - 130 + kostka.y * pomer, W_HEIGHT - 115);
        glEnd();
    }
}

static void
draw_target(
    float pomer_x,
    float pomer_y)
{
    int target_x;
    int target_y;

    target_x = souradnice_x(pix.target_x_SL, pomer_x);
    target_y = souradnice_y(pix.target_y_SL, pomer_y);

    /* Red target
     *
     *   |
     * --+--
     *   |
     */
    if ((pix.target_CB) && (rozsah(target_x, target_y, VWIDTH, VHEIGHT))) {
        glColor3f(1.0, 0.0, 0.0);
        glBegin(GL_LINES);
        glVertex2i(target_x, target_y - 40);
        glVertex2i(target_x, target_y + 40);
        glVertex2i(target_x - 40, target_y);
        glVertex2i(target_x + 40, target_y);
        glEnd();
    }

    if (pix.show_cut_xy_CB) {
        /* Red target in cut X
         *
         *   |
         *   |
         *   |
         */
        if ((pix.target_CB) && (rozsah(target_x, target_y, VWIDTH, VHEIGHT))) {
            glColor3f(1.0, 0.0, 0.0);
            glBegin(GL_LINES);
            glVertex2i(target_x, W_HEIGHT);
            glVertex2i(target_x, W_HEIGHT - W_REZ);
            glEnd();
        }

        /* Violet cursor slith in cut X
         *
         *   |
         *   |
         *   |
         */
        if (pix.show_cursor_slith_CB) {
            glColor3f(1.0, 0.4, 1.0);
            glBegin(GL_LINES);
            glVertex2i(pix.slith_begin_SL, W_HEIGHT);
            glVertex2i(pix.slith_begin_SL, W_HEIGHT - W_REZ);
            glVertex2i(pix.slith_end_SL, W_HEIGHT);
            glVertex2i(pix.slith_end_SL, W_HEIGHT - W_REZ);
            glEnd();
        }

        /* Red target in cut Y
         *
         * -----
         */
        if ((pix.target_CB) && (rozsah(target_x, target_y, VWIDTH, VHEIGHT))) {
            glColor3f(1.0, 0.0, 0.0);
            glBegin(GL_LINES);
            glVertex2i(W_WIDTH, target_y);
            glVertex2i(W_WIDTH - W_REZ, target_y);
            glEnd();
        }
    }
}

static void
draw_value_in_quadrant(
    void)
{
    int x;
    int y;

    if ((pix.show_value_in_quadrant_CB) && (pix.zoom_CB)) {
        /* A */
        glColor3f(1.0, 1.0, 1.0);
        glBegin(GL_QUADS);
        glVertex2i(W_MEZERA, W_MEZERA);
        glVertex2i(W_MEZERA, W_MEZERA + 20);
        glVertex2i(W_MEZERA + 40, W_MEZERA + 20);
        glVertex2i(W_MEZERA + 40, W_MEZERA);
        glEnd();

        glColor3f(0.0, 0.0, 0.0);
        printInt(W_MEZERA + 5, W_MEZERA + 15, kvadrant.a);

        /* B */
        glColor3f(1.0, 1.0, 1.0);
        x = VWIDTH / 2 + W_MEZERA;
        glBegin(GL_QUADS);
        glVertex2i(x, W_MEZERA);
        glVertex2i(x, W_MEZERA + 20);
        glVertex2i(x + 40, W_MEZERA + 20);
        glVertex2i(x + 40, W_MEZERA);
        glEnd();

        glColor3f(0.0, 0.0, 0.0);
        printInt(x + 5, W_MEZERA + 15, kvadrant.b);

        /* C */
        glColor3f(1.0, 1.0, 1.0);
        y = VHEIGHT / 2;
        glBegin(GL_QUADS);
        glVertex2i(W_MEZERA, y);
        glVertex2i(W_MEZERA, y + 20);
        glVertex2i(W_MEZERA + 40, y + 20);
        glVertex2i(W_MEZERA + 40, y);
        glEnd();

        glColor3f(0.0, 0.0, 0.0);
        printInt(W_MEZERA + 5, y + 15, kvadrant.c);

        /* D */
        glColor3f(1.0, 1.0, 1.0);
        x = VWIDTH / 2 + W_MEZERA;
        y = VHEIGHT / 2 + W_MEZERA;
        glBegin(GL_QUADS);
        glVertex2i(x, y);
        glVertex2i(x, y + 20);
        glVertex2i(x + 40, y + 20);
        glVertex2i(x + 40, y);
        glEnd();

        glColor3f(0.0, 0.0, 0.0);
        printInt(x + 5, y + 15, kvadrant.d);
    }
}

static void
draw_background(
    void)
{
    /* Background right bottom square */
    glColor3f(0.9, 0.9, 0.9);
    glBegin(GL_QUADS);
    glVertex2i(W_WIDTH - 255, W_HEIGHT);
    glVertex2i(W_WIDTH - 255, W_HEIGHT - 255);
    glVertex2i(W_WIDTH, W_HEIGHT - 255);
    glVertex2i(W_WIDTH, W_HEIGHT);
    glEnd();

    if (pix.show_cut_xy_CB) {
        /* Background cut X */
        glColor3f(1.0, 1.0, 1.0);
        glBegin(GL_QUADS);
        glVertex2i(0, VHEIGHT + W_MEZERA);
        glVertex2i(0, VHEIGHT + W_MEZERA + W_REZ);
        glVertex2i(VWIDTH, VHEIGHT + W_MEZERA + W_REZ);
        glVertex2i(VWIDTH, VHEIGHT + W_MEZERA);
        glEnd();

        /* Background cut Y */
        glColor3f(1.0, 1.0, 1.0);
        glBegin(GL_QUADS);
        glVertex2i(VWIDTH + W_MEZERA, 0);
        glVertex2i(VWIDTH + W_MEZERA, VHEIGHT);
        glVertex2i(VWIDTH + W_MEZERA + W_REZ, VHEIGHT);
        glVertex2i(VWIDTH + W_MEZERA + W_REZ, 0);
        glEnd();
    }
}

static void
draw_cursor_zoom_cut(
    int sirka,
    int vyska)
{
    int x;
    int y;

    if (pix.show_cut_xy_CB) {
        /* Blue cursor zoom in cut X
         *
         * --------------------------
         *
         * --------------------------
         */
        if ((pix.show_cursor_zoom_cut_CB) && (!pix.zoom_cut_CB)) {
            glColor3f(0.0, 0.0, 1.0);
            glBegin(GL_LINES);
            y = W_HEIGHT - pix.zoom_cut_x_SL;
            glVertex2i(0, y);
            glVertex2i(VWIDTH, y);
            glVertex2i(0, y - pix.zoom_cut_x_size_SL);
            glVertex2i(VWIDTH, y - pix.zoom_cut_x_size_SL);
            glEnd();
        }

        /* Blue cursor zoom in cut Y
         *
         * | |
         * | |
         * | |
         * | |
         * | |
         */
        if ((pix.show_cursor_zoom_cut_CB) && (!pix.zoom_cut_CB)) {
            glColor3f(0.0, 0.0, 1.0);
            glBegin(GL_LINES);
            x = W_WIDTH - pix.zoom_cut_y_SL;
            glVertex2i(x, vyska);
            glVertex2i(x, vyska - sirka);
            glVertex2i(x - pix.zoom_cut_y_size_SL, vyska);
            glVertex2i(x - pix.zoom_cut_y_size_SL, vyska - sirka);
            glEnd();
        }
    }
}

static void
draw_cut_x(
    float pomer_x,
    float pomer_y,
    float *p_pomer_cut_x,
    int *p_star_x)
{
    unsigned char pixel;
    unsigned char pixel_max = 255;
    unsigned char slith_y_begin = 0;
    unsigned char slith_y_end = 0;
    int i;
    int j;
    int index = 0;
    int y_cut;
    int soucet;
    int pocatek = 0;
    int sirka = VWIDTH;
    int vyska = VHEIGHT;
    int sirka_pole = VWIDTH;
    int slith_width;
    int slith_height;
    float y;
    float y_max = 0;
    float fitgaus_count[FITGSL_NUM_COEFF];

    if (pix.show_cut_xy_CB) {
        /* Black cut X */
        if (pix.zoom_cut_CB) {
            pixel_max = pix.zoom_cut_x_size_SL + pix.zoom_cut_x_SL - 1;
            *p_pomer_cut_x = 255.0 / pix.zoom_cut_x_size_SL;
            glScalef(1.0, *p_pomer_cut_x, 0.0);
        }

        if (pix.zoom_CB) {
            sirka = pix.zoom_size_SL;
            sirka_pole = sirka;
            glScalef(pomer_x, 1.0, 0.0);

            if ((pix.cut_all_lines_zoom_CB) || (!pix.histogram_CB)) {
                vyska = pix.zoom_size_SL;
            }
            else {
                vyska = VHEIGHT;
                sirka_pole = VWIDTH;
                pocatek = pix.zoom_x_SL - (pix.zoom_size_SL / 2);
                if (pix.average_SL == 1) {
                    buffer.p_pole = AKTUALNI_SNIMEK_SM;
                }
                else {
                    buffer.p_pole = buffer.pole;
                }
            }
        }
        else {
            pomer_x = 1.0;
        }

        slith_width = pix.slith_end_SL - pix.slith_begin_SL;

        glColor3f(0.0, 0.0, 0.0);
        glBegin(GL_LINE_STRIP);
        for (i = 0; i < sirka; i++) {
            if ((pix.histogram_CB) && (pix.cut_x_size_SL > 0)) {
                soucet = 0;
                //for (j = 0; j < vyska; ++j) {
                for (j = pix.cut_x_SL - pix.cut_x_size_SL;
                     j < pix.cut_x_SL + pix.cut_x_size_SL; ++j) {
                    soucet +=
                        buffer.p_pole[(vyska - j) * sirka_pole + i + pocatek];
                }
                pixel = soucet / (2 * pix.cut_x_size_SL);
            }
            else {
                pixel =
                    buffer.p_pole[(vyska - (int) (pix.cut_x_SL / pomer_x)) *
                                  sirka + i];
            }

            if (pixel > pixel_max) {
                pixel = pixel_max;
            }

            y_cut = W_HEIGHT / *p_pomer_cut_x - pixel;
            if (pix.zoom_cut_CB) {
                y_cut += pix.zoom_cut_x_SL;
            }

            if (pix.show_fit_gaus_xy_CB) {
                if (i == pix.slith_begin_SL) {
                    slith_y_begin = pixel;
                }
                else if (i == pix.slith_end_SL) {
                    slith_y_end = pixel;
                }

                if ((i <= pix.slith_begin_SL) || (i >= pix.slith_end_SL)) {
                    // TODO: pridat hlidani mezi pole
                    fitgaus_data_x->pt[index].x = (float) i;
                    fitgaus_data_x->pt[index].y = (float) pixel;
                    ++index;
                }
            }
            glVertex2f(i, y_cut);
        }
        glEnd();

        fitgaus_data_x->n = index;

        if (pix.show_fit_gaus_xy_CB) {
            /*
             *    slith_width = slith_x_end - slith_x_begin
             *    slith_height = slith_y_end - slith_y_begin
             *
             *    P[x; y]
             *    Py = (Px / (slith_width / 100)) * (slith_height / 100)
             */

            // DBG osetrit totozne y_begin a y_end!!!

            if (slith_y_begin > slith_y_end) {
                pixel = slith_y_begin;
                slith_y_begin = slith_y_end;
                slith_y_end = pixel;
            }

            slith_height = slith_y_end - slith_y_begin;

            //printf("*************************************************************\n");
            //printf("slith_width = %i\n", slith_width);
            //printf("slith_height = %i\n", slith_height);
            //printf("slith_y_begin = %i\n", slith_y_begin);
            //printf("slith_y_end = %i\n", slith_y_end);

            //for (i = pix.slith_begin_SL + 1; i < pix.slith_end_SL; i++) {
            //    pixel =
            //        ((i -
            //          pix.slith_begin_SL) / (slith_width / 100.0)) *
            //        (slith_height / 100.0);
            //    pixel += slith_y_begin;

            //    // TODO: pridat hlidani mezi pole
            //    fitgaus_data_x->pt[i].x = (float) i;
            //    fitgaus_data_x->pt[i].y = (float) pixel;

            //    //printf("x = %i, y = %i\n", i - pix.slith_begin_SL, pixel);
            //}
        }

        if (pix.show_fit_gaus_xy_CB) {
            fitgsl_lm(fitgaus_data_x, fitgaus_count, 0);

            fwhm = 2.35482 * fitgaus_count[FITGSL_W_INDEX];
            //printf (" FWHM = %f\n", fwhm);

            glColor3f(0.0, 0.0, 1.0);
            glBegin(GL_LINE_STRIP);
            for (i = 0; i < sirka; i++) {
                y = fitgsl_fx(fitgaus_count[FITGSL_B_INDEX],
                              fitgaus_count[FITGSL_P_INDEX],
                              fitgaus_count[FITGSL_C_INDEX],
                              fitgaus_count[FITGSL_W_INDEX], (float) i);

                if (y_max < y) {
                    y_max = y;
                    intenzity_star = y_max;
                    *p_star_x = i;
                }

                if (y > pixel_max) {
                    y = pixel_max;
                }

                y_cut = W_HEIGHT / *p_pomer_cut_x - y;
                if (pix.zoom_cut_CB) {
                    y_cut += pix.zoom_cut_x_SL;
                }
                glVertex2f(i, y_cut);
            }
            glEnd();
        }

        if (pix.zoom_cut_CB) {
            glScalef(1.0, 1.0 / *p_pomer_cut_x, 0.0);
        }

        if (pix.show_fit_gaus_xy_CB) {
            glColor3f(0.0, 0.0, 1.0);
            glBegin(GL_LINES);
            glVertex2i(*p_star_x, W_HEIGHT);
            glVertex2i(*p_star_x, W_HEIGHT - W_REZ);
            glEnd();
        }

        if (pix.zoom_CB) {
            glScalef(1.0 / pomer_x, 1.0, 0.0);
        }

        if (pix.show_fit_gaus_xy_CB) {
            printFWHM(10, W_HEIGHT - W_REZ + 20, fwhm);
        }
    }
}

static void
draw_cut_y(
    float pomer_x,
    float pomer_y,
    float pomer_cut_x,
    int *p_star_y)
{
    unsigned char pixel;
    unsigned char pixel_max = 255;
    int i;
    int j;
    int x;
    int soucet;
    int pocatek = 0;
    int sirka = VWIDTH;
    int vyska = VHEIGHT;
    int sirka_pole = VWIDTH;
    int cut_y_slith_begin = pix.cut_y_SL - (pix.cut_y_slith_size_SL / 2);
    int cut_y_slith_end = pix.cut_y_SL + (pix.cut_y_slith_size_SL / 2);
    int count;
    float y;
    float y_max;
    float pomer_cut_y = 1.0;
    float fitgaus_count[FITGSL_NUM_COEFF];

    if (pix.show_cut_xy_CB) {
        /* Black cut Y */
        if (pix.zoom_cut_CB) {
            pixel_max = pix.zoom_cut_y_size_SL + pix.zoom_cut_y_SL - 1;
            pomer_cut_y = 255.0 / pix.zoom_cut_y_size_SL;
            glScalef(pomer_cut_y, 1.0, 0.0);
        }
        else {
            pixel_max = 255;
            pomer_cut_y = 1.0;
        }

        if (pix.zoom_CB) {
            sirka = pix.zoom_size_SL;
            glScalef(1.0, pomer_y, 0.0);

            if ((pix.cut_all_lines_zoom_CB) || (!pix.histogram_CB)) {
                vyska = pix.zoom_size_SL;
            }
            else {
                vyska = pix.zoom_size_SL;
                sirka_pole = VWIDTH;
                pocatek = pix.zoom_y_SL - (pix.zoom_size_SL / 2);
                if (pix.average_SL == 1) {
                    buffer.p_pole = AKTUALNI_SNIMEK_SM;
                }
                else {
                    buffer.p_pole = buffer.pole;
                }
            }
        }
        else {
            pomer_y = 1.0;
        }

        glColor3f(0.0, 0.0, 0.0);
        glBegin(GL_LINE_STRIP);
        for (i = 0; i < vyska; i++) {
            if ((pix.histogram_CB) && (pix.cut_y_size_SL > 0)) {
                soucet = 0;
                count = 0;
                //for (j = 0; j < sirka_pole; ++j) {
                for (j = pix.cut_y_SL - pix.cut_y_size_SL;
                     j < pix.cut_y_SL + pix.cut_y_size_SL; ++j) {
                    if ((pix.cut_y_slith_size_SL == 0)
                        || (j < cut_y_slith_begin) || (j > cut_y_slith_end)) {
                        soucet +=
                            buffer.p_pole[(i + pocatek) * sirka_pole + j];
                        ++count;
                    }
                }

                if (count > 0) {
                    pixel = soucet / count;
                }
            }
            else {
                pixel =
                    buffer.p_pole[i * sirka + (int) (pix.cut_y_SL / pomer_y)];
            }

            if (pixel > pixel_max) {
                pixel = pixel_max;
            }

            if (pix.show_fit_gaus_xy_CB) {
                // TODO: pridat hlidani mezi pole
                fitgaus_data_y->pt[i].x = (float) i;
                fitgaus_data_y->pt[i].y = (float) pixel;
            }

            x = W_WIDTH / pomer_cut_y - pixel;
            if (pix.zoom_cut_CB) {
                x += pix.zoom_cut_y_SL;
            }
            glVertex2f(x, vyska - i);
        }
        glEnd();

        if (pix.show_fit_gaus_xy_CB) {
            fitgsl_lm(fitgaus_data_y, fitgaus_count, 0);

            y_max = 0;

            glColor3f(0.0, 0.0, 1.0);
            glBegin(GL_LINE_STRIP);
            for (i = 0; i < vyska; i++) {
                y = fitgsl_fx(fitgaus_count[FITGSL_B_INDEX],
                              fitgaus_count[FITGSL_P_INDEX],
                              fitgaus_count[FITGSL_C_INDEX],
                              fitgaus_count[FITGSL_W_INDEX], (float) i);

                if (y_max < y) {
                    y_max = y;
                    *p_star_y = vyska - i;
                }

                if (y > pixel_max) {
                    y = pixel_max;
                }

                x = W_WIDTH / pomer_cut_y - y;
                if (pix.zoom_cut_CB) {
                    x += pix.zoom_cut_y_SL;
                }
                glVertex2f(x, vyska - i);
            }
            glEnd();

            if (pix.zoom_cut_CB) {
                glScalef(1.0 / pomer_cut_y, 1.0, 0.0);
            }

            glColor3f(0.0, 0.0, 1.0);
            glBegin(GL_LINES);
            glVertex2i(W_WIDTH, *p_star_y);
            glVertex2i(W_WIDTH - W_REZ, *p_star_y);
            glEnd();
        }
    }
}

static void
draw_fit_gaus_target(
    int star_x,
    int star_y,
    float pomer_x)
{
    if (pix.show_fit_gaus_xy_CB) {
        if (pix.zoom_CB) {
            glScalef(pomer_x, 1.0, 0.0);
        }

        glColor3f(0.0, 0.0, 1.0);
        glBegin(GL_LINES);
        glVertex2i(star_x, star_y - 20);
        glVertex2i(star_x, star_y + 20);
        glVertex2i(star_x - 20, star_y);
        glVertex2i(star_x + 20, star_y);
        glEnd();
    }
}

static void
send_TSGS(
    int star_x,
    int star_y)
{
    static time_t allowed_offset_time = 0;

    int i;
    float offset_x;
    float offset_y;
    float sec_in_pix = pix.sec_in_pix_SB * 0.001;
    float deviation = fabsf(pix.permissible_deviation_SB * 0.01);
    float allowed_offset = fabsf(pix.allowed_offset_SB);

    offset_x = pix.target_x_SL - star_x;
    offset_y = pix.target_y_SL - star_y;
    offset_x = -offset_x;

    /*
     *          0.2mm = 0.6s
     *    slith width = 25pix
     *           1pix = 0.6 / 25 = 0.024s
     */
    offset_x *= sec_in_pix;
    offset_y *= sec_in_pix;

    for (i = (TSGC_HISTORY_LEN - 1); i >= 0; --i) {
        strcpy(tsgc_history[i], tsgc_history[i - 1]);
    }

    if ((pix.permissible_deviation_SB != 0) &&
        (fabsf(offset_x) < deviation) && (fabsf(offset_y) < deviation)) {
        snprintf(tsgc_history[0], COMMAND_MAX,
                 "Permisible deviation %.02f %.02f", offset_x, offset_y);
        return;
    }

    if ((pix.min_intenzity_star_SB != 0) &&
        (intenzity_star < pix.min_intenzity_star_SB)) {
        snprintf(tsgc_history[0], COMMAND_MAX, "Min intenzity star > %i",
                 intenzity_star);
        return;
    }

    if ((pix.max_fwhm_star_SB != 0) && (fwhm > pix.max_fwhm_star_SB)) {
        snprintf(tsgc_history[0], COMMAND_MAX, "Max FWHM star < %0.1f", fwhm);
        return;
    }

    if (allowed_offset_time == 0) {
        allowed_offset_time = time(NULL) + pix.timeout_allowed_offset_SB;
    }

    if ((pix.allowed_offset_SB != 0) &&
        ((fabsf(offset_x) >= allowed_offset)
         || (fabsf(offset_y) >= allowed_offset))
        && (allowed_offset_time > time(NULL))) {
        snprintf(tsgc_history[0], COMMAND_MAX, "Allowed offset %.02f %.02f",
                 offset_x, offset_y);
        return;
    }

    allowed_offset_time = 0;

    //printf("star[%i; %i]\n", star_x, star_y);
    //printf("target[%i; %i]\n", pix.target_x_SL, pix.target_y_SL);
    //printf("TSGC %f %f\n", offset_x, offset_y);
    //printf("sec_in_pix %f\n", sec_in_pix);

    snprintf(command, COMMAND_MAX, "./ps_tsgc %.02f %.02f &", offset_x,
             offset_y);
    system(command);
    syslog(LOG_INFO, "Execute: %s", command);

    snprintf(tsgc_history[0], COMMAND_MAX, "TSGC %.02f %.02f", offset_x,
             offset_y);

    //printf("max_fwhm_star_SB = %i\n", pix.max_fwhm_star_SB);
    //printf("min_intenzity_star_SB = %i\n", pix.min_intenzity_star_SB);
    //printf("corr_at_error_SB = %i\n", pix.corr_at_error_SB);
    //printf("allowed_offset_SB = %i\n", pix.allowed_offset_SB);
    //printf("timeout_allowed_offset_SB = %i\n", pix.timeout_allowed_offset_SB);
}

/**
  *
 */
static void
onDisplay(
    void)
{
    static time_t tsgc_time = 0;

    int star_x = 0;
    int star_y = 0;
    int sirka = VWIDTH;
    int vyska = VHEIGHT;
    float pomer_x = 1.0;
    float pomer_y = 1.0;
    float pomer_cut_x = 1.0;

    lock(pix_semid, 0);
    pix = *p_pix_sm;
    p_pix_sm->zmena_pru = 0;
    unlock(pix_semid, 0);

    kostka.max = pix.max_count_step_SB;

    lock(ccbuf_semid, 0);
    if (p_ccbuf_sm->snimek == 0) {
        snimek = SNIMKU - 1;
    }
    else {
        snimek = p_ccbuf_sm->snimek - 1;
    }
    unlock(ccbuf_semid, 0);

    if (pix.calibration == 1) {
        //tty_write (0, 3, 0);
        //tty_write (1, 3, 0);
        //tty_write (0, 1, pix.kostka_start_x);
        //tty_write (1, 1, pix.kostka_start_y);

        kostka.x = 0;
        kostka.y = 0;

        lock(pix_semid, 0);
        p_pix_sm->calibration = 0;
        unlock(pix_semid, 0);
    }

    pomer_x = (float) VWIDTH / pix.zoom_size_SL;
    pomer_y = (float) VHEIGHT / pix.zoom_size_SL;

    buffer.p_pole = pointer();

    glClear(GL_COLOR_BUFFER_BIT);
    glPushMatrix();

    /* Brightness */
    glPixelTransferf(GL_RED_BIAS, pix.brightness_SL / 255.0f);
    glPixelTransferf(GL_GREEN_BIAS, pix.brightness_SL / 255.0f);
    glPixelTransferf(GL_BLUE_BIAS, pix.brightness_SL / 255.0f);

    /* Contrast */
    glPixelTransferf(GL_RED_SCALE, pix.contrast_SL / 255.0f);
    glPixelTransferf(GL_GREEN_SCALE, pix.contrast_SL / 255.0f);
    glPixelTransferf(GL_BLUE_SCALE, pix.contrast_SL / 255.0f);

    /* DEBUG */
    if (pix.show_cut_xy_CB) {
        glRasterPos2i(0, VHEIGHT_ZOOM);
    }
    else {
        glRasterPos2i(0, window.height);
    }

    if (pix.zoom_CB == 1) {
        if (pix.show_cut_xy_CB) {
            glPixelZoom((float) VWIDTH_ZOOM / pix.zoom_size_SL,
                        (float) VHEIGHT_ZOOM / pix.zoom_size_SL);
        }
        else {
            glPixelZoom((float) window.width / pix.zoom_size_SL,
                        (float) window.height / pix.zoom_size_SL);
        }
        glDrawPixels(pix.zoom_size_SL, pix.zoom_size_SL, GL_LUMINANCE,
                     GL_UNSIGNED_BYTE, buffer.p_pole);
    }
    else {
        if (pix.show_cut_xy_CB) {
            glPixelZoom((float) VWIDTH_ZOOM / VWIDTH,
                        (float) VHEIGHT_ZOOM / VHEIGHT);
        }
        else {
            glPixelZoom((float) window.width / VWIDTH,
                        (float) window.height / VHEIGHT);
        }
        glDrawPixels(VWIDTH, VHEIGHT, GL_LUMINANCE, GL_UNSIGNED_BYTE,
                     buffer.p_pole);
    }

    if (pix.show_cut_xy_CB) {
        glScalef((float) window.width / W_WIDTH,
                 (float) window.height / W_HEIGHT, 0.0);
    }
    else {
        glScalef((float) window.width / VWIDTH,
                 (float) window.height / VHEIGHT, 0.0);
    }
    /* DEBUG */

    draw_cursor_cut_xy();
    draw_cursor_slith();
    draw_blacken_faulty_pix(pomer_x, pomer_y);
    draw_faulty_pix(pomer_x, pomer_y);
    draw_cursor_faulty_pix(pomer_x, pomer_y);
    draw_background();
    draw_cube_position();
    draw_target(pomer_x, pomer_y);
    draw_value_in_quadrant();
    draw_cursor_zoom_cut(sirka, vyska);
    draw_cut_x(pomer_x, pomer_y, &pomer_cut_x, &star_x);
    draw_cut_y(pomer_x, pomer_y, pomer_cut_x, &star_y);
    draw_fit_gaus_target(star_x, star_y, pomer_x);

    if ((pix.automation_pointing_TB) && (tsgc_time <= time(NULL))) {
        tsgc_time = time(NULL) + pix.timeout_tsgc_SB;
        //printf("timeout_tsgc_SB = %i, time = %i\n", pix.timeout_tsgc_SB, tsgc_time);
        send_TSGS(star_x, star_y);
    }
    else if (!pix.automation_pointing_TB) {
        tsgc_time = 0;
    }

    // DBG
    {
        int i;

        for (i = 0; i < TSGC_HISTORY_LEN; ++i) {
            if (i == 0) {
                glColor3f(1.0, 0.0, 0.0);
            }
            else if (i == 1) {
                glColor3f(0.0, 0.0, 1.0);
            }

            printString(W_WIDTH - W_REZ + 10,
                        W_HEIGHT - W_REZ + (20 * (i + 1)), tsgc_history[i]);
        }
    }

    glPopMatrix();
    glutSwapBuffers();
}

/**
  * 
 */
static void
run_long_option(
    const char *option,
    const char *optarg)
{
    if (option == "width") {
        window.width = atoi(optarg);
    }
    else if (option == "height") {
        window.height = atoi(optarg);
    }
    else if (option == "x") {
        window.x = atoi(optarg);
    }
    else if (option == "y") {
        window.y = atoi(optarg);
    }
}

/**
  * MAIN
 */
int
main(
    int argc,
    char *argv[])
{
    struct sigaction act;

    openlog("opso", LOG_PID | LOG_CONS, LOG_LOCAL0);
    syslog(LOG_INFO, "RUN pointace");

    if (!file_exist(PID_PATH "opso.pid")) {
        syslog(LOG_ERR, "ERROR Nebezi opso");
        exit(EXIT_FAILURE);
    }

    window.width = VWIDTH + W_MEZERA + W_REZ;
    window.height = VHEIGHT + W_MEZERA + W_REZ;
    window.x = 0;
    window.y = 0;

    while (1) {
        int o;
        int option_index = 0;
        static struct option long_options[] = {
            {"width", 1, 0, 0},
            {"height", 1, 0, 0},
            {"x", 1, 0, 0},
            {"y", 1, 0, 0},
            {0, 0, 0, 0}
        };

        o = getopt_long(argc, argv, "", long_options, &option_index);
        if (o == -1) {
            break;
        }

        switch (o) {
        case 0:
            run_long_option(long_options[option_index].name, optarg);
            break;

        default:
            syslog(LOG_INFO, "getopt vrtila 0%o", o);
            break;
        }
    }

    act.sa_handler = konec;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGTERM, &act, 0);

    write_pid("pointace", getpid());
    ipc_create();
    //tty_open ();

    buffer.snimek = 0;
    buffer.naplnen = 0;

    kostka.x = 0;
    kostka.y = 0;
    //kostka.max = 200;
    kostka.max = 50;

    fitgaus_data_x = fitgsl_alloc_data(VWIDTH);
    fitgaus_data_y = fitgsl_alloc_data(VHEIGHT);

    memset(tsgc_history, 0, sizeof(tsgc_history));

    system("./ovladani.sh &");

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
    glutInitWindowSize(window.width, window.height);
    glutInitWindowPosition(window.x, window.y);
    glutCreateWindow("OPSO pointing");
    glClearColor(0.7, 0.7, 0.7, 0.7);
    glShadeModel(GL_FLAT);
    glutDisplayFunc(onDisplay);
    glutReshapeFunc(onReshape);
    glutIdleFunc(onIdle);
    glutCloseFunc(onClose);
    glutMainLoop();

    return 0;
}
