#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define PS_TSGC_VERSION "0.9.0"

#define PS_SERVER_IP   "192.168.193.20"
#define PS_SERVER_PORT 2002

#define PS_IP_MAX      32
#define PS_RECVBUF_MAX 64
#define PS_CMD_MAX     64
#define PS_RECORD_MAX  1024

#define OFFSET_MAX 1023

static int ps_sockfd;
static char buffer[PS_RECVBUF_MAX];

static void
ps_help(
    char *name)
{
    printf("Usage: %s [OFFSET_X OFFSET_Y]\n", name);
    printf("  -h, --help                 Print this help and exit\n");
    printf
        ("      --version              Display version information and exit\n\n");
    printf("Default parametrs: %s:%i\n\n", PS_SERVER_IP, PS_SERVER_PORT);
}

static void
ps_version(
    char *name)
{
    printf("%s %s (%s, %s)\n", name, PS_TSGC_VERSION, __TIME__, __DATE__);
    printf("Written by Jan Fuchs <fuky@sunstel.asu.cas.cz>\n\n");

    printf("Copyright (C) 2008 Free Software Foundation, Inc.\n");
    printf
        ("This is free software; see the source for copying conditions.  There is NO\n");
    printf
        ("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
}

static int
ps_socket_read(
    char *p_buffer,
    int buffer_size)
{
    int result;
    fd_set rfd;
    char recvbuf[PS_RECVBUF_MAX];
    struct timeval timeout;

    memset(p_buffer, 0, buffer_size);

    FD_ZERO(&rfd);
    FD_SET(ps_sockfd, &rfd);

    timeout.tv_sec = 5;
    timeout.tv_usec = 0;

    result = select(FD_SETSIZE, &rfd, (fd_set *) 0, (fd_set *) 0, &timeout);

    if (result > 0) {
        result = recv(ps_sockfd, recvbuf, PS_RECVBUF_MAX, 0);
        if (result > 0) {
            if (result < buffer_size) {
                buffer_size = result;
            }
            strncpy(p_buffer, recvbuf, buffer_size);
            return 0;
        }
    }

    return -1;
}

static int
ps_socket_write(
    char *p_msg)
{
    int msg_len;
    int count;
    struct timeval timeout;
    fd_set wfd;

    msg_len = strlen(p_msg);

    FD_ZERO(&wfd);
    FD_SET(ps_sockfd, &wfd);

    timeout.tv_sec = 5;
    timeout.tv_usec = 0;

    if (select(FD_SETSIZE, (fd_set *) 0, &wfd, (fd_set *) 0, &timeout) <= 0) {
        perror("select() failure");
        return -1;
    }

    if ((count = send(ps_sockfd, p_msg, msg_len, 0)) != msg_len) {
        perror("send() failure");
        return -1;
    }

    return 0;
}

static int
ps_socket_create(
    char *p_ip,
    int port)
{
    unsigned long flag_ioctl;
    fd_set wfd;
    struct timeval timeout;
    static struct sockaddr_in server_address;

    bzero((char *) &server_address, sizeof(server_address));
    server_address.sin_family = AF_INET;
    server_address.sin_addr.s_addr = inet_addr(p_ip);
    server_address.sin_port = htons(port);

    if ((ps_sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
        perror("socket() failure");
        return -1;
    }

    flag_ioctl = 1;
    if (ioctl(ps_sockfd, FIONBIO, &flag_ioctl) == -1) { // SET NONBLOCK
        perror("ioctl(FIONBIO) failure");
        return -1;
    }

    if (connect
        (ps_sockfd, (struct sockaddr *) &server_address,
         sizeof(server_address)) == -1) {
        if (errno != EINPROGRESS) {
            perror("connect() failure");
            return -1;
        }
    }

    FD_ZERO(&wfd);
    FD_SET(ps_sockfd, &wfd);

    timeout.tv_sec = 5;
    timeout.tv_usec = 0;

    if (select(FD_SETSIZE, (fd_set *) 0, &wfd, (fd_set *) 0, &timeout) <= 0) {
        perror("select() failure");
        return -1;
    }

    return 0;
}

int
main(
    int argc,
    char *argv[])
{
    char ip[PS_IP_MAX];
    int port = PS_SERVER_PORT;
    char offset_x[OFFSET_MAX + 1];
    char offset_y[OFFSET_MAX + 1];

    strcpy(ip, PS_SERVER_IP);

    switch (argc) {
    case 2:
        if (!strcmp(argv[1], "--version")) {
            ps_version(argv[0]);
        }
        else {
            ps_help(argv[0]);
        }
        exit(EXIT_SUCCESS);
        break;

    case 3:
        strcpy(offset_x, argv[1]);
        strcpy(offset_y, argv[2]);
        break;

    default:
        exit(EXIT_SUCCESS);
        break;
    }

    if (ps_socket_create(ip, port) == -1) {
        close(ps_sockfd);
        exit(EXIT_FAILURE);
    }

    if (ps_socket_write("GLLG 123\n") == -1) {
        exit(EXIT_FAILURE);
    }

    ps_socket_read(buffer, PS_RECVBUF_MAX);
    if (buffer[0] != '1') {
        printf("Authentification failure: %s\n", buffer);
        exit(EXIT_FAILURE);
    }

    sprintf(buffer, "TSGC %s %s\n", offset_x, offset_y);
    ps_socket_write(buffer);
    ps_socket_read(buffer, PS_RECVBUF_MAX);

    close(ps_sockfd);
    exit(EXIT_SUCCESS);
}
