#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define TELESCOPE_SERVER_IP   "192.168.193.20"
#define TELESCOPE_SERVER_PORT 2004

#define TELESCOPE_IP_MAX      32
#define TELESCOPE_RECVBUF_MAX 64
#define TELESCOPE_CMD_MAX     64

static int telescope_sockfd;

int
telescope_socket_read(
    char *p_buffer,
    int buffer_size)
{
    int result;
    fd_set rfd;
    char recvbuf[TELESCOPE_RECVBUF_MAX];
    struct timeval timeout;

    memset(p_buffer, 0, buffer_size);

    FD_ZERO(&rfd);
    FD_SET(telescope_sockfd, &rfd);

    timeout.tv_sec = 5;
    timeout.tv_usec = 0;

    result = select(FD_SETSIZE, &rfd, (fd_set *) 0, (fd_set *) 0, &timeout);

    if (result > 0) {
        result = recv(telescope_sockfd, recvbuf, TELESCOPE_RECVBUF_MAX, 0);
        if (result > 0) {
            if (result < buffer_size) {
                buffer_size = result;
            }
            strncpy(p_buffer, recvbuf, buffer_size);
            return 0;
        }
    }

    return -1;
}

int
telescope_socket_write(
    char *p_msg)
{
    int msg_len;
    int count;
    struct timeval timeout;
    fd_set wfd;

    msg_len = strlen(p_msg);

    FD_ZERO(&wfd);
    FD_SET(telescope_sockfd, &wfd);

    timeout.tv_sec = 5;
    timeout.tv_usec = 0;

    if (select(FD_SETSIZE, (fd_set *) 0, &wfd, (fd_set *) 0, &timeout) <= 0) {
        perror("select() failure");
        return -1;
    }

    if ((count = send(telescope_sockfd, p_msg, msg_len, 0)) != msg_len) {
        perror("send() failure");
        return -1;
    }

    return 0;
}

int
telescope_socket_create(
    char *p_ip,
    int port)
{
    unsigned long flag_ioctl;
    fd_set wfd;
    struct timeval timeout;
    static struct sockaddr_in server_address;

    bzero((char *) &server_address, sizeof(server_address));
    server_address.sin_family = AF_INET;
    server_address.sin_addr.s_addr = inet_addr(p_ip);
    server_address.sin_port = htons(port);

    if ((telescope_sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
        perror("socket() failure");
        return -1;
    }

    flag_ioctl = 1;
    if (ioctl(telescope_sockfd, FIONBIO, &flag_ioctl) == -1) {  // SET NONBLOCK
        perror("ioctl(FIONBIO) failure");
        return -1;
    }

    if (connect
        (telescope_sockfd, (struct sockaddr *) &server_address,
         sizeof(server_address)) == -1) {
        if (errno != EINPROGRESS) {
            perror("connect() failure");
            return -1;
        }
    }

    FD_ZERO(&wfd);
    FD_SET(telescope_sockfd, &wfd);

    timeout.tv_sec = 5;
    timeout.tv_usec = 0;

    if (select(FD_SETSIZE, (fd_set *) 0, &wfd, (fd_set *) 0, &timeout) <= 0) {
        perror("select() failure");
        return -1;
    }

    return 0;
}

int
main(
    int argc,
    char *argv[])
{
    char buffer[TELESCOPE_RECVBUF_MAX];
    char cmd[TELESCOPE_CMD_MAX];
    char ip[TELESCOPE_IP_MAX];
    int i;
    int len;
    int port = TELESCOPE_SERVER_PORT;

    strcpy(ip, TELESCOPE_SERVER_IP);

    if (telescope_socket_create(ip, port) == -1) {
        close(telescope_sockfd);
        exit(EXIT_FAILURE);
    }

    for (i = 0; i <= TELESCOPE_GLME_MAX; ++i) {
        sprintf(cmd, "GLME %i\n", i);
        if (telescope_socket_write(cmd) == -1) {
            close(telescope_sockfd);
            exit(EXIT_FAILURE);
        }

        telescope_socket_read(buffer, TELESCOPE_RECVBUF_MAX);
        len = strlen(buffer) - 2;
        if (len > 0) {
            strncpy(telescope_glme[i], buffer, len);
        }
    }
}

int
telescope_socket_close(
    )
{
    close(telescope_sockfd);
}
